"""
This module acts as a simple wrapper around dcm2niix, it takes all of the same arguments as dcm2niix but does a little
bit of extra work to conform the output nifti and json from dcm2niix to the PET BIDS specification. Additionally, but
optionally, this module can collect blood or physiological data/metadata from spreadsheet files if the path of that
spreadsheet file as well as a python module/script written to interpret it are provided in addition to relevant dcm2niix
commands.

For more details see the CLI portion of this module or the documentation for the main class Dcm2niix4PET


| *Authors: Anthony Galassi*
| *Copyright OpenNeuroPET team*
"""

import pathlib
import sys
import textwrap
from json_maj.main import JsonMAJ
from platform import system
import subprocess
import pandas as pd
from os.path import join
from os import listdir, walk, environ
from pathlib import Path
import json
import pydicom
import re
from tempfile import TemporaryDirectory
import shutil
import argparse
import importlib

try:
    import helper_functions
    import is_pet
    from update_json_pet_file import (
        check_json,
        update_json_with_dicom_value,
        update_json_with_dicom_value_cli,
        get_radionuclide,
        check_meta_radio_inputs,
        metadata_dictionaries,
        get_metadata_from_spreadsheet,
    )
    from telemetry import (
        send_telemetry,
        count_input_files,
        telemetry_enabled,
        count_output_files,
    )
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions
    import pypet2bids.is_pet as is_pet
    from pypet2bids.update_json_pet_file import (
        check_json,
        update_json_with_dicom_value,
        update_json_with_dicom_value_cli,
        get_radionuclide,
        check_meta_radio_inputs,
        metadata_dictionaries,
        get_metadata_from_spreadsheet,
    )
    from pypet2bids.telemetry import (
        send_telemetry,
        count_input_files,
        telemetry_enabled,
        count_output_files,
    )

logger = helper_functions.logger("pypet2bids")

module_folder = Path(__file__).parent.resolve()
python_folder = module_folder.parent
pet2bids_folder = python_folder.parent
metadata_folder = join(pet2bids_folder, "metadata")

# check to see if config file exists
home_dir = Path.home()
pypet2bids_config = home_dir / ".pet2bidsconfig"
if pypet2bids_config.exists():
    # check to see if the template json var is set and valid
    default_metadata_json = helper_functions.check_pet2bids_config(
        "DEFAULT_METADATA_JSON"
    )
    if default_metadata_json and Path(str(default_metadata_json)).exists():
        # do nothing
        pass
    elif default_metadata_json and not Path(str(default_metadata_json)).exists():
        # Copy template to the specified path
        try:
            # Ensure the parent directory exists before copying
            Path(default_metadata_json).parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(
                Path(metadata_folder) / "template_json.json", default_metadata_json
            )
        except FileNotFoundError:
            # Fallback to module folder template
            Path(default_metadata_json).parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(module_folder / "template_json.json", default_metadata_json)
    else:
        # No DEFAULT_METADATA_JSON set in config, do nothing
        pass
else:
    # if it doesn't exist use the default one included in this library
    helper_functions.modify_config_file(
        "DEFAULT_METADATA_JSON", module_folder / "template_json.json"
    )
    # set the default metadata json to the template json included in this library in our environment so
    # we don't trip up retrieval of the default metadata json later
    environ["DEFAULT_METADATA_JSON"] = str(module_folder / "template_json.json")
    default_metadata_json = module_folder / "template_json.json"


def dicom_datetime_to_dcm2niix_time(dicom=None, date="", time=""):
    """
    Dcm2niix provides the option of outputting the scan data and time into the .nii and .json filename at the time of
    conversion if '%t' is provided following the '-f' flag. The result is the addition of a date time string of the
    format. This function similarly generates the same datetime string from a dicom header.

    :param dicom: pydicom.dataset.FileDataset object or a path to a dicom
    :param date: a given date, used in conjunction with time to supply a date time
    :param time: a given time, used in conjunction with date

    :return: a datetime string that corresponds to the converted filenames from dcm2niix when used with the `-f %t` flag
    """
    parsed_time = ""
    parsed_date = ""
    if dicom:
        if type(dicom) is pydicom.dataset.FileDataset:
            # do nothing
            pass
        elif type(dicom) is str:
            try:
                dicom_path = Path(dicom)
                dicom = pydicom.dcmread(dicom_path)
            except TypeError:
                raise TypeError(
                    f"dicom {dicom} must be either a pydicom.dataset.FileDataSet object or a "
                    f"valid path to a dicom file"
                )

        parsed_date = dicom.StudyDate
        parsed_time = str(round(float(dicom.StudyTime)))

    elif date and time:
        parsed_date = date
        parsed_time = str(round(float(time)))

    if len(parsed_time) < 6:
        zeros_to_pad = 6 - len(parsed_time)
        parsed_time = zeros_to_pad * "0" + parsed_time

    return parsed_date + parsed_time


def collect_date_time_from_file_name(file_name):
    """
    Collects the date and time from a nifti or a json produced by dcm2niix when dcm2niix is run with the options
    %p_%i_%t_%s. This datetime us used to match a dicom header object to the resultant file. E.G. if there are missing
    BIDS fields in the json produced by dcm2niix it's hopeful that the dicom header may contain the missing info.
    :param file_name: name of the file to extract the date time info from, this should be a json output file from
    dcm2niix
    :return: a date and time object
    """
    date_time_string = re.search(r"(?!\_)[0-9]{14}(?=\_)", file_name)
    if date_time_string:
        date = date_time_string[0][0:8]
        time = date_time_string[0][8:]
    else:
        raise Exception(f"Unable to parse date_time string from filename: {file_name}")

    return date, time


class Dcm2niix4PET:
    def __init__(
        self,
        image_folder,
        destination_path=None,
        metadata_path=None,
        metadata_translation_script=None,
        additional_arguments={},
        dcm2niix_options="",
        file_format="%p_%i_%t_%s",
        silent=False,
        tempdir_location=None,
        ezbids=False,
        ignore_dcm2niix_errors=False,
    ):
        """
        This class is a simple wrapper for dcm2niix and contains methods to do the following in order:
            - Convert a set of dicoms into .nii and .json sidecar files
            - Inspect the .json sidecar files for any missing BIDS PET fields or values
            - If there are missing BIDS PET fields or values this class will attempt to extract them from the dicom
            header, a metadata file using a metadata translation script, and lastly from user supplied key pair
            arguments.

        # class is instantiated:
        converter = Dcm2niix4PET(...)
        # then conversion is run by calling run_dcm2niix
        converter.run_dcm2niix()

        Conversion is performed in a temporary directory to make matching dicom headers to dcm2niix output files easier
        (and to avoid leaving intermediary files persisting on disc). After which, these files are then moved the
        destination directory.

        :param image_folder: folder containing a single series/session of dicoms
        :param destination_path: destination path for dcm2niix output nii and json files
        :param metadata_path: path to excel, csv, or text file with PET metadata (radioligand, blood, etc etc)
        :param metadata_translation_script: python file to extract and transform data contained in the metadata_path
        :param file_format: the file format that we want dcm2niix to use, by default %p_%i_%t_%s
        %p -> protocol
        %i -> ID of patient
        %t -> time
        %s -> series number
        :param additional_arguments: user supplied key value pairs, E.g. TimeZero=\"12:12:12\", InjectedRadioactivity=1
        this key value pair will overwrite any fields in the dcm2niix produced nifti sidecar.json as it is assumed that
        the user knows more about converting their data than the heuristics within dcm2niix, this library, or even the
        dicom header
        :param tempdir_location: user supplied base location for temporary directory (override system default)
        :param silent: silence missing sidecar metadata messages, default is False and very verbose
        :param tempdir_location: location to create the temporary directory, for use on constrained systems
        """

        # check to see if dcm2niix is installed
        self.blood_json = None
        self.blood_tsv = None
        self.telemetry_data = {}
        self.ezbids = ezbids
        self.dcm2niix_path = self.check_for_dcm2niix()
        if not self.dcm2niix_path:
            logger.error(
                "dcm2niix not found, this module depends on it for conversions, exiting."
            )
            sys.exit(1)
        self.ignore_dcm2niix_errors = ignore_dcm2niix_errors
        # check for the version of dcm2niix
        minimum_version = "v1.0.20220720"
        version_string = subprocess.run([self.dcm2niix_path, "-v"], capture_output=True)
        version = re.search(r"v[0-9].[0-9].{8}[0-9]", str(version_string.stdout))

        if version:
            # compare with minimum version
            if version[0] < minimum_version:
                logger.warning(
                    f"Minimum version {minimum_version} of dcm2niix is recommended, found "
                    f"installed version {version[0]} at {self.dcm2niix_path}."
                )

        # check if user provided a custom tempdir location
        self.tempdir_location = tempdir_location
        self.image_folder = Path(image_folder)
        self.destination_folder = None

        # if we're provided an entire file path just us that no matter what, we're assuming the user knows what they
        # are doing in that case
        self.full_file_path_given = False

        for part in Path(destination_path).parts:
            if ".nii" in part or ".nii.gz" in part:
                self.full_file_path_given = True
                self.destination_folder = Path(destination_path).parent
                # replace .nii and .nii.gz
                self.destination_path = Path(
                    str(destination_path).replace(".nii", "").replace(".gz", "")
                )
                break

        # replace the suffix in the destination path with '' if a non-nifti full file path is give
        if Path(destination_path).suffix:
            self.full_file_path_given = True
            self.destination_folder = Path(destination_path).parent
            self.destination_path = Path(destination_path).with_suffix("")

        if not self.full_file_path_given:
            if not destination_path:
                self.destination_path = self.image_folder
                self.destination_folder = self.image_folder
            else:
                self.destination_folder = Path(destination_path)
                self.destination_path = self.destination_folder

        # extract PET filename parts from destination path if given
        self.subject_id = helper_functions.collect_bids_part(
            "sub", str(self.destination_path)
        )
        self.session_id = helper_functions.collect_bids_part(
            "ses", str(self.destination_path)
        )
        self.task = helper_functions.collect_bids_part(
            "task", str(self.destination_path)
        )
        self.tracer = helper_functions.collect_bids_part(
            "trc", str(self.destination_path)
        )
        self.reconstruction_method = helper_functions.collect_bids_part(
            "rec", str(self.destination_path)
        )
        self.run_id = helper_functions.collect_bids_part(
            "run", str(self.destination_path)
        )

        self.file_name_slug = None

        # we keep track of PET metadata in this spreadsheet metadata_dict, that includes nifti, _blood.json, and
        # _blood.tsv data
        self.spreadsheet_metadata = {
            "nifti_json": {},
            "blood_json": {},
            "blood_tsv": {},
        }
        self.dicom_headers = self.extract_dicom_headers()
        # we consider values stored in a default JSON file to be additional arguments, we load those
        # values first and then overwrite them with any user supplied values

        # check to make sure the dicom is a PET dicom
        modalities = [dicom.Modality for dicom in self.dicom_headers.values()]
        non_pet_dicom = [modality for modality in modalities if modality != "PT"]

        if len(non_pet_dicom) > 0:
            logger.warning(
                f"Non PET dicoms found in {self.image_folder}. Modalities f{non_pet_dicom} detected"
            )

        # load config file
        default_json_path = helper_functions.check_pet2bids_config(
            "DEFAULT_METADATA_JSON"
        )
        if default_json_path and Path(default_json_path).exists():
            with open(default_json_path, "r") as json_file:
                try:
                    self.spreadsheet_metadata.update(json.load(json_file))
                except json.decoder.JSONDecodeError:
                    logger.warning(
                        f"Unable to load default metadata json file at {default_json_path}, skipping."
                    )

        self.additional_arguments = additional_arguments

        # if there's a spreadsheet and if there's a provided python script use it to manipulate the data in the
        # spreadsheet
        if metadata_path and metadata_translation_script:
            self.metadata_path = Path(metadata_path)
            self.metadata_translation_script = Path(metadata_translation_script)

            if (
                self.metadata_path.exists()
                and self.metadata_translation_script.exists()
            ):
                # load the spreadsheet into a dataframe
                self.extract_metadata()
                # next we use the loaded python script to extract the information we need
                self.load_spread_sheet_data()
        elif metadata_path and not metadata_translation_script or metadata_path == "":
            self.metadata_path = Path(metadata_path)
            if not self.spreadsheet_metadata.get("nifti_json", None):
                self.spreadsheet_metadata["nifti_json"] = {}

            load_spreadsheet_data = get_metadata_from_spreadsheet(
                metadata_path=metadata_path,
                image_folder=self.image_folder,
                image_header_dict=self.dicom_headers[next(iter(self.dicom_headers))],
                **self.additional_arguments,
            )

            self.spreadsheet_metadata["nifti_json"].update(
                load_spreadsheet_data["nifti_json"]
            )
            self.spreadsheet_metadata["blood_tsv"].update(
                load_spreadsheet_data["blood_tsv"]
            )
            self.spreadsheet_metadata["blood_json"].update(
                load_spreadsheet_data["blood_json"]
            )

            # we default to these options, a user may supply their own combination, this will allow full customization of all
        # but the file arguments for dcm2niix
        if dcm2niix_options:
            # Filter out -f flag and its value from custom options since file format is handled separately
            options_list = dcm2niix_options.split()
            filtered_options = []
            i = 0
            while i < len(options_list):
                if options_list[i] == "-f" and i + 1 < len(options_list):
                    # Skip -f and its value
                    i += 2
                else:
                    filtered_options.append(options_list[i])
                    i += 1
            self.dcm2niix_options = " ".join(filtered_options)
        else:
            # Check for dcm2niix options in config file and environment variable
            config_options = helper_functions.check_pet2bids_config(
                "PET2BIDS_DCM2NIIX_OPTIONS"
            )
            env_options = environ.get("PET2BIDS_DCM2NIIX_OPTIONS")

            # Priority: command line > environment variable > config file > default
            if env_options:
                self.dcm2niix_options = env_options
                logger.info(
                    f"Using dcm2niix options from environment variable: {env_options}"
                )
            elif config_options:
                self.dcm2niix_options = config_options
                logger.info(
                    f"Using dcm2niix options from config file: {config_options}"
                )
            else:
                self.dcm2niix_options = "-b y -w 1 -z y"
                logger.debug("Using default dcm2niix options: -b y -w 1 -z y")
        self.file_format = file_format
        # we may want to include additional information to the sidecar, tsv, or json files generated after conversion
        # this variable stores the mapping between output files and a single dicom header used to generate those files
        # to access the dicom header information use the key in self.headers_to_files to access that specific header
        # in self.dicom_headers
        self.headers_to_files = {}
        # if silent is set to True output warnings aren't displayed to stdout/stderr
        self.silent = silent

    @staticmethod
    def check_posix():
        check = subprocess.run(
            "dcm2niix -h",
            shell=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        if check.returncode == 0:
            dcm2niix_path = (
                subprocess.run("which dcm2niix", shell=True, capture_output=True)
                .stdout.decode("utf-8")
                .strip()
            )
            # check to see if dcm2niix is set in the config file, default to that if it's the case and alert user
            set_in_config = helper_functions.check_pet2bids_config()
            if set_in_config:
                logger.warning(
                    f"dcm2niix found on system path, but dcm2niix path is also set in ~/.pet2bidsconfig."
                    f" Defaulting to dcm2niix path set in config at {set_in_config}"
                )
                dcm2niix_path = set_in_config

        else:
            dcm2niix_path = helper_functions.check_pet2bids_config()

        if not dcm2niix_path:
            pkged = "https://github.com/rordenlab/dcm2niix/releases"
            instructions = "https://github.com/rordenlab/dcm2niix#install"
            no_dcm2niix = f"""Unable to locate Dcm2niix on your system $PATH or using the path specified in 
                        $HOME/.pypet2bidsconfig. Installation instructions for dcm2niix can be found here 
                        {instructions} 
                        and packaged versions can be found at 
                        {pkged}
                        Alternatively, you can set the path to dcm2niix in the config file at $HOME/.pet2bidsconfig
                        using the command dcm2niix4pet --set-dcm2niix-path."""
            logger.error(no_dcm2niix)
            dcm2niix_path = None

        return dcm2niix_path

    def check_for_dcm2niix(self):
        """
        Just checks for dcm2niix using the system shell, returns 0 if dcm2niix is present.

        :return: status code of the command dcm2niix -h
        """

        if system().lower() != "windows":
            dcm2niix_path = self.check_posix()
            # fall back and check the config file if it's not on the path
            if not dcm2niix_path:
                dcm2niix_path = helper_functions.check_pet2bids_config("DCM2NIIX_PATH")
        elif system().lower() == "windows":
            dcm2niix_path = helper_functions.check_pet2bids_config("DCM2NIIX_PATH")
        else:
            dcm2niix_path = None

        return dcm2niix_path

    def extract_dicom_headers(self, depth=1):
        """
        Opening up files till a dicom is located, then extracting any header information
        to be used during and after the conversion process. This includes patient/subject id,
        as well any additional frame or metadata that's required for conversion.

        :param depth: the number of dicoms to collect per folder, defaults to 1 as it assumes a single sessions worth of
                     dicoms is included per folder.
        :return: dicom header information to self.subject_id and/or self.dicom_header_data
        """
        n = 0
        dicom_headers = {}
        for root, dirs, files in walk(self.image_folder):
            for f in files:
                if n >= depth:
                    break
                try:
                    dicom_path = Path(join(root, f))
                    dicom_header = pydicom.dcmread(dicom_path, stop_before_pixels=True)
                    # collect subject/patient id if none is supplied
                    if self.subject_id is None:
                        self.subject_id = dicom_header.PatientID

                    dicom_headers[dicom_path.name] = dicom_header
                    n += 1

                except pydicom.errors.InvalidDicomError:
                    pass

        return dicom_headers

    def run_dcm2niix(self):
        """
        This runs dcm2niix and uses the other methods within this class to supplement the sidecar json's produced as
        dcm2niix output.

        :return: the path to the output of dcm2niix and the modified sidecar jsons
        """
        if self.file_format:
            file_format_args = f"-f {self.file_format}"
        else:
            file_format_args = ""
        with TemporaryDirectory(dir=self.tempdir_location) as tempdir:
            tempdir_pathlike = Path(tempdir)
            self.tempdir_location = tempdir_pathlike
            # people use screwy paths, we do this before running dcm2niix to account for that
            image_folder = helper_functions.sanitize_bad_path(self.image_folder)
            cmd = f"{self.dcm2niix_path} {self.dcm2niix_options} {file_format_args} -o {tempdir_pathlike} {image_folder}"
            convert = subprocess.run(cmd, shell=True, capture_output=True)
            self.telemetry_data["dcm2niix"] = {
                "returncode": convert.returncode,
            }
            self.telemetry_data.update(count_output_files(self.tempdir_location))

            if (
                convert.returncode != 0
                or "error" in convert.stderr.decode("utf-8").lower()
            ) and not self.ignore_dcm2niix_errors:
                print(
                    "Check output .nii files, dcm2iix returned these errors during conversion:"
                )
                # raise error if missing images is found
                if "missing images" in convert.stderr.decode("utf8").lower():
                    error_message = convert.stderr.decode("utf8").replace("Error:", "")
                    error_message = f"{error_message}for dicoms in {image_folder}"
                    raise FileNotFoundError(error_message)
                if (
                    bytes("Skipping existing file name", "utf-8") not in convert.stdout
                    or convert.stderr
                ):
                    print(convert.stderr.decode("utf-8"))
                elif (
                    convert.returncode != 0
                    and bytes("Error: Check sorted order", "utf-8") in convert.stdout
                    or convert.stderr
                ):
                    print(
                        "Possible error with frame order, is this a phillips dicom set?"
                    )
                    print(convert.stdout)
                    print(convert.stderr)

            # collect contents of the tempdir
            files_created_by_dcm2niix = [
                join(tempdir_pathlike, file) for file in listdir(tempdir_pathlike)
            ]

            # check to see if there are any modalities present that aren't PET
            remove_these_non_pet_files = []
            for json_file in [f for f in files_created_by_dcm2niix if ".json" in f]:
                with open(json_file, "r") as infile:
                    json_data = json.load(infile)
                    if json_data.get("Modality") != "PT":
                        logger.warning(
                            f"Non PET modality found in {json_file}, skipping. Files relating to this json and it's"
                            f"nifti will not be included at the destination folder {self.destination_folder}"
                        )
                        remove_these_non_pet_files.append(json_file)
                        remove_these_non_pet_files.append(
                            json_file.replace(".json", ".nii.gz")
                        )
                        remove_these_non_pet_files.append(
                            json_file.replace(".json", ".nii")
                        )

            # if there are non PET files remove them from files_created_by_dcm2niix
            if remove_these_non_pet_files:
                for file in remove_these_non_pet_files:
                    try:
                        files_created_by_dcm2niix.remove(file)
                    except ValueError:
                        pass

            # make sure destination path exists if not try creating it.
            try:
                if self.destination_path.exists():
                    pass
                elif not self.destination_path.exists() and self.full_file_path_given:
                    self.destination_path.parent.mkdir(parents=True, exist_ok=True)
                else:
                    self.destination_path.mkdir(parents=True, exist_ok=True)
            except FileExistsError:
                pass

            # iterate through created files to supplement sidecar jsons
            for created in files_created_by_dcm2niix:
                created_path = Path(created)
                if created_path.suffix == ".json":
                    # we want to pair up the headers to the files created in the output directory in case
                    # dcm2niix has created files from multiple sessions
                    matched_dicoms_and_headers = self.match_dicom_header_to_file(
                        destination_path=tempdir_pathlike
                    )

                    # we check to see what's missing from our recommended and required jsons by gathering the
                    # output of check_json silently
                    if self.additional_arguments:
                        check_for_missing = check_json(
                            created_path,
                            silent=True,
                            spreadsheet_metadata=self.spreadsheet_metadata,
                            **self.additional_arguments,
                        )
                    else:
                        check_for_missing = check_json(
                            created_path,
                            silent=True,
                            spreadsheet_metadata=self.spreadsheet_metadata,
                        )

                    # we do our best to extra information from the dicom header and insert these values
                    # into the sidecar json

                    # first do a reverse lookup of the key the json corresponds to
                    lookup = [
                        key
                        for key, value in matched_dicoms_and_headers.items()
                        if str(created_path) in value
                    ]
                    if lookup:
                        dicom_header = self.dicom_headers[lookup[0]]

                        update_json_with_dicom_value(
                            created_path,
                            check_for_missing,
                            dicom_header,
                            dicom2bids_json=metadata_dictionaries["dicom2bids"],
                            ezbids=self.ezbids,
                            **self.additional_arguments,
                        )

                    # if we have entities in our metadata spreadsheet that we've used we update
                    if self.spreadsheet_metadata.get("nifti_json", None):
                        update_json = JsonMAJ(
                            json_path=str(created),
                            update_values=self.spreadsheet_metadata["nifti_json"],
                            bids_null=True,
                        )
                        update_json.update()

                    # check to see if frame duration is a single value, if so convert it to list
                    update_json = JsonMAJ(json_path=str(created), bids_null=True)

                    # there are some additional updates that depend on some PET BIDS logic that we do next, since these
                    # updates depend on both information provided via the sidecar json and/or information provided via
                    # additional arguments we run this step after updating the sidecar with those additional user
                    # arguments

                    sidecar_json = JsonMAJ(
                        json_path=str(created),
                        bids_null=True,
                        update_values=self.additional_arguments,
                    )  # load all supplied and now written sidecar data in

                    sidecar_json.update()

                    check_metadata_radio_inputs = check_meta_radio_inputs(
                        sidecar_json.json_data
                    )  # run logic

                    sidecar_json.update(
                        check_metadata_radio_inputs
                    )  # update sidecar json with results of logic

                    # should be list/array types in the json
                    should_be_array = [
                        "FrameDuration",
                        "ScatterFraction",
                        "FrameTimesStart",
                        "DecayCorrectionFactor",
                        "ReconFilterSize",
                    ]

                    for should in should_be_array:
                        should_value = update_json.get(should)
                        if should_value and type(should_value) is not list:
                            update_json.update({should: [should_value]})

                    # next we check to see if any of the additional user supplied arguments (kwargs) correspond to
                    # any of the missing tags in our sidecars
                    if self.additional_arguments:
                        update_json = JsonMAJ(
                            json_path=str(created),
                            update_values=self.additional_arguments,
                            bids_null=True,
                        )
                        update_json.update()

                    # check to see if convolution kernel is present
                    sidecar_json = JsonMAJ(json_path=str(created), bids_null=True)
                    if sidecar_json.get("ConvolutionKernel"):
                        if sidecar_json.get("ReconFilterType") and sidecar_json.get(
                            "ReconFilterSize"
                        ):
                            sidecar_json.remove("ConvolutionKernel")
                        else:
                            # collect filter size
                            recon_filter_size = ""
                            if re.search(
                                r"\d+.\d+", sidecar_json.get("ConvolutionKernel")
                            ):
                                try:
                                    recon_filter_size = re.search(
                                        r"\d+.\d*",
                                        sidecar_json.get("ConvolutionKernel"),
                                    )[0]
                                    recon_filter_size = float(recon_filter_size)
                                except ValueError:
                                    # If float conversion fails, try splitting and take first part
                                    match_str = re.search(
                                        r"\d+.\d*",
                                        sidecar_json.get("ConvolutionKernel"),
                                    )[0]
                                    recon_filter_size = float(match_str.split()[0])
                                sidecar_json.update(
                                    {"ReconFilterSize": float(recon_filter_size)}
                                )
                            # collect just the filter type by popping out the filter size if it exists
                            recon_filter_type = re.sub(
                                str(recon_filter_size),
                                "",
                                sidecar_json.get("ConvolutionKernel"),
                            )
                            # further sanitize the recon filter type string
                            recon_filter_type = re.sub(
                                r"[^a-zA-Z0-9]", " ", recon_filter_type
                            )
                            recon_filter_type = re.sub(r" +", " ", recon_filter_type)

                            # update the json
                            sidecar_json.update({"ReconFilterType": recon_filter_type})
                            # remove non bids field
                            sidecar_json.remove("ConvolutionKernel")

                    # check the input args again as our logic is applied after parsing user inputs
                    if self.additional_arguments:
                        recon_filter_user_input = {
                            "ReconFilterSize": self.additional_arguments.get(
                                "ReconFilterSize", None
                            ),
                            "ReconFilterType": self.additional_arguments.get(
                                "ReconFilterType", None
                            ),
                        }
                        for key, value in recon_filter_user_input.items():
                            if value:
                                sidecar_json.update({key: value})
                    else:
                        pass

                    # tag json with additional conversion software
                    conversion_software = sidecar_json.get("ConversionSoftware")
                    conversion_software_version = sidecar_json.get(
                        "ConversionSoftwareVersion"
                    )

                    sidecar_json.update(
                        {"ConversionSoftware": [conversion_software, "pypet2bids"]}
                    )
                    sidecar_json.update(
                        {
                            "ConversionSoftwareVersion": [
                                conversion_software_version,
                                helper_functions.get_version(),
                            ]
                        }
                    )

                    # if this looks familiar, that's because it is, we re-run this to override any changes
                    # made by this software as the input provided by the user is "the correct input"
                    sidecar_json.update(self.spreadsheet_metadata.get("nifti_json", {}))
                    sidecar_json.update(self.additional_arguments)

                    # set ModeOfAdministration to lower case
                    if sidecar_json.get("ModeOfAdministration"):
                        sidecar_json.update(
                            {
                                "ModeOfAdministration": sidecar_json.get(
                                    "ModeOfAdministration"
                                ).lower()
                            }
                        )

                    # this is mostly for ezBIDS, but it helps us to make better use of the series description that
                    # dcm2niix generates by default for PET imaging
                    collect_these_fields = {
                        "ProtocolName": "",
                        "SeriesDescription": "",
                        "TracerName": "trc",
                        "InjectedRadioactivity": "",
                        "InjectedRadioactivityUnits": "",
                        "ReconMethodName": "rec",
                        "TimeZero": "",
                    }
                    collection_of_fields = {}
                    for field, entity_string in collect_these_fields.items():
                        if sidecar_json.get(field):
                            # if there's a shortened entity string for the field use that
                            if entity_string != "":
                                collection_of_fields[entity_string] = sidecar_json.get(
                                    field
                                )
                            else:
                                collection_of_fields[field] = sidecar_json.get(field)

                    if self.session_id:
                        collection_of_fields["ses"] = self.session_id

                    if self.ezbids:
                        hash_string = helper_functions.hash_fields(
                            **collection_of_fields
                        )
                        sidecar_json.update({"SeriesDescription": hash_string})

                # if there's a subject id rename the output file to use it
                if self.subject_id:
                    if "nii.gz" in created_path.name:
                        suffix = ".nii.gz"
                    else:
                        suffix = created_path.suffix
                    if self.session_id:
                        session_id = "_" + self.session_id
                    else:
                        session_id = ""

                    if self.task:
                        task = "_" + self.task
                    else:
                        task = ""

                    if self.tracer:
                        trc = "_" + self.tracer
                    else:
                        trc = ""

                    if self.reconstruction_method:
                        rec = "_" + self.reconstruction_method
                    else:
                        rec = ""

                    if self.run_id:
                        run = "_" + self.run_id
                    else:
                        run = ""

                    if self.full_file_path_given:
                        new_path = self.destination_path.with_suffix(suffix)
                        self.destination_folder = self.destination_path.parent
                    else:
                        new_path = self.destination_path / Path(
                            self.subject_id
                            + session_id
                            + task
                            + trc
                            + rec
                            + run
                            + "_pet"
                            + suffix
                        )

                    try:
                        new_path.parent.mkdir(parents=True, exist_ok=True)
                    except FileExistsError:
                        pass

                elif not self.subject_id:
                    new_path = Path(join(self.destination_path, created_path.name))

                self.new_file_name_with_entities = new_path

                shutil.move(src=created, dst=new_path)

            return self.destination_path

    def post_dcm2niix(self):
        # TODO add logic to handle blood tsv recording manual vs automatic
        # for now we will just assume that if the user supplied a blood tsv then it is manual
        recording_entity = "_recording-manual"

        if "_pet" in self.new_file_name_with_entities.name:
            if (
                self.new_file_name_with_entities.suffix == ".gz"
                and len(self.new_file_name_with_entities.suffixes) > 1
            ):
                self.new_file_name_with_entities = (
                    self.new_file_name_with_entities.with_suffix("").with_suffix("")
                )

            blood_file_name = self.new_file_name_with_entities.stem.replace(
                "_pet", recording_entity + "_blood"
            )
        else:
            blood_file_name = (
                self.new_file_name_with_entities.stem + recording_entity + "_blood"
            )

        if self.spreadsheet_metadata.get("blood_tsv", {}) != {}:
            blood_tsv_data = self.spreadsheet_metadata.get("blood_tsv")
            if type(blood_tsv_data) is pd.DataFrame or type(blood_tsv_data) is dict:
                if type(blood_tsv_data) is dict:
                    blood_tsv_data = pd.DataFrame(blood_tsv_data)
                # write out blood_tsv using pandas csv write
                blood_tsv_data.to_csv(
                    join(self.destination_folder, blood_file_name + ".tsv"),
                    sep="\t",
                    index=False,
                )

            elif type(blood_tsv_data) is str:
                # write out with write
                with open(
                    join(self.destination_folder, blood_file_name + ".tsv"), "w"
                ) as outfile:
                    outfile.writelines(blood_tsv_data)
            else:
                raise (
                    f"blood_tsv dictionary is incorrect type {type(blood_tsv_data)}, must be type: "
                    f"pandas.DataFrame or str\nCheck return type of translate_metadata in "
                    f"{self.metadata_translation_script}"
                )

        # if there's blood data in the tsv then write out the sidecar file too
        if (
            self.spreadsheet_metadata.get("blood_json", {}) != {}
            and self.spreadsheet_metadata.get("blood_tsv", {}) != {}
        ):
            blood_json_data = self.spreadsheet_metadata.get("blood_json")
            if type(blood_json_data) is dict:
                # write out to file with json dump
                pass
            elif type(blood_json_data) is str:
                # write out to file with json dumps
                blood_json_data = json.loads(blood_json_data)
            else:
                raise (
                    f"blood_json dictionary is incorrect type {type(blood_json_data)}, must be type: dict or str"
                    f"pandas.DataFrame or str\nCheck return type of translate_metadata in "
                    f"{self.metadata_translation_script}"
                )

            with open(
                join(self.destination_folder, blood_file_name + ".json"), "w"
            ) as outfile:
                json.dump(blood_json_data, outfile, indent=4)

    def convert(self):
        # check the size of out the output folder
        before_output_files = {}
        self.run_dcm2niix()
        self.post_dcm2niix()

        # if telemetry isn't disabled we send a telemetry event to the pypet2bids server
        if telemetry_enabled:
            # count the number of files
            self.telemetry_data.update(count_input_files(self.image_folder))
            # record if a blood tsv and json file were created
            if self.spreadsheet_metadata.get("blood_tsv", {}) != {}:
                self.telemetry_data["blood_tsv"] = True
            else:
                self.telemetry_data["blood_tsv"] = False
            # record if a metadata spreadsheet was used
            if helper_functions.collect_spreadsheets(self.metadata_path):
                self.telemetry_data["metadata_spreadsheet_used"] = True
            else:
                self.telemetry_data["metadata_spreadsheet_used"] = False

            self.telemetry_data["InputType"] = "DICOM"

            self.telemetry_data["returncode"] = 0

            send_telemetry(self.telemetry_data)

    def match_dicom_header_to_file(self, destination_path=None):
        """
        Matches a dicom header to a nifti or json file produced by dcm2niix, this is run after dcm2niix converts the
        input dicoms into nifti's and json's.

        :param destination_path: the path dcm2niix generated files are placed at, collected during class instantiation

        :return: a dictionary of headers matched to nifti and json file names
        """
        if not destination_path:
            destination_path = self.destination_path
        # first collect all the files in the output directory
        output_files = [
            join(destination_path, output_file)
            for output_file in listdir(destination_path)
        ]

        # create empty dictionary to store pairings
        headers_to_files = {}

        # collect study date and time from header
        for each in self.dicom_headers:
            header_study_date = self.dicom_headers[each].StudyDate
            header_acquisition_time = str(
                round(float(self.dicom_headers[each].StudyTime))
            )
            if len(header_acquisition_time) < 6:
                header_acquisition_time = (
                    6 - len(header_acquisition_time)
                ) * "0" + header_acquisition_time

            header_date_time = dicom_datetime_to_dcm2niix_time(
                date=header_study_date, time=header_acquisition_time
            )

            for output_file in output_files:
                if header_date_time in output_file:
                    try:
                        headers_to_files[each].append(output_file)
                    except KeyError:
                        headers_to_files[each] = [output_file]
        return headers_to_files

    def extract_metadata(self):
        """
        Opens up a metadata file and reads it into a pandas dataframe
        :return: a pd dataframe object
        """
        # collect metadata from spreadsheet
        metadata_extension = Path(self.metadata_path).suffix
        self.open_meta_data(metadata_extension)

    def open_meta_data(self, extension):
        """
        Opens a text metadata file with the pandas method most appropriate for doing so based on the metadata
        file's extension.
        :param extension: The extension of the file
        :return: a pandas dataframe representation of the spreadsheet/metadatafile
        """
        methods = {"excel": pd.read_excel, "csv": pd.read_csv}

        if "xls" in extension:
            proper_method = "excel"
        else:
            proper_method = extension

        try:
            use_me_to_read = methods.get(proper_method, None)
            self.metadata_dataframe = use_me_to_read(self.metadata_path)
        except IOError as err:
            raise err(f"Problem opening {self.metadata_path}")

    def load_spread_sheet_data(self):
        text_file_data = {}
        if self.metadata_translation_script:
            try:
                # this is where the goofiness happens, we allow the user to create their own custom script to manipulate
                # data from their particular spreadsheet wherever that file is located.
                spec = importlib.util.spec_from_file_location(
                    "metadata_translation_script", self.metadata_translation_script
                )
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                text_file_data = module.translate_metadata(self.metadata_dataframe)
            except AttributeError as err:
                print(f"Unable to locate metadata_translation_script")

            self.spreadsheet_metadata["blood_tsv"] = text_file_data.get("blood_tsv", {})
            self.spreadsheet_metadata["blood_json"] = text_file_data.get(
                "blood_json", {}
            )
            self.spreadsheet_metadata["nifti_json"] = text_file_data.get(
                "nifti_json", {}
            )


epilog = textwrap.dedent(
    """
    
    example usage:
    
    dcm2niix4pet folder_with_pet_dicoms/ --destination-path sub-ValidBidSSubject/pet # the simplest conversion
    dcm2niix4pet folder_with_pet_dicoms/ --destination-path sub-ValidBidsSubject/pet --metadata-path metadata.xlsx \
    # use with an input spreadsheet
    dcm2niix4pet folder_with_pet_dicoms/ --destination-path sub-ValidBidsSubject/pet --dcm2niix-options -v y -w 1 -z y \
    # use with custom dcm2niix options
    dcm2niix4pet --set-dcm2niix-options '-v y -w 1 -z y' \
    # set default dcm2niix options in config file
    
"""
)


def cli():
    """
    Collects arguments used to initiate a Dcm2niix4PET class, collects the following arguments from the user.

    :param folder: folder containing imaging data, no flag required
    :param -m, --metadata-path: path to PET metadata spreadsheet
    :param -t, --translation-script-path: path to script used to extract information from metadata spreadsheet
    :param -d, --destination-path: path to place outputfiles post conversion from dicom to nifti + json
    :return: arguments collected from argument parser
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=epilog,
        description="Given a set of PET dicoms and additional metadata dcm2niix converts "
        "them to BIDS compliant nifti (using dcm2niix), json, and tsv files.",
    )
    parser.add_argument(
        "folder", nargs="?", type=str, help="Folder path containing imaging data"
    )
    parser.add_argument(
        "--metadata-path",
        "-m",
        type=str,
        default=None,
        const="",
        nargs="?",
        help="Path to metadata file for scan",
    )
    parser.add_argument(
        "--translation-script-path",
        "-t",
        default=None,
        help="Path to a script written to extract and transform metadata from a spreadsheet to BIDS"
        + " compliant text files (tsv and json)",
    )
    parser.add_argument(
        "--destination-path",
        "-d",
        type=str,
        default=None,
        help="Destination path to send converted imaging and metadata files to. If subject id and "
        "session id is included in the path files created by dcm2niix4pet will be named as such. "
        "e.g. sub-NDAR123/ses-ABCD/pet will yield fields named sub-NDAR123_ses-ABCD_*. If "
        + "omitted defaults to using the path supplied to folder path. If destination path "
        + "doesn't exist an attempt to create it will be made.",
        required=False,
    )
    parser.add_argument(
        "--tempdir",
        type=str,
        default=None,
        help="User-specified tempdir location (overrides default system tempfile default)",
        required=False,
    )
    parser.add_argument(
        "--kwargs",
        "-k",
        nargs="*",
        action=helper_functions.ParseKwargs,
        default={},
        help="Include additional values in the nifti sidecar json or override values extracted from "
        'the supplied nifti. e.g. including `--kwargs TimeZero="12:12:12"` would override the '
        "calculated TimeZero. Any number of additional arguments can be supplied after --kwargs "
        "e.g. `--kwargs BidsVariable1=1 BidsVariable2=2` etc etc."
        "Note: the value portion of the argument (right side of the equal's sign) should always"
        'be surrounded by double quotes BidsVarQuoted="[0, 1 , 3]"',
    )
    parser.add_argument(
        "--silent",
        "-s",
        action="store_true",
        default=False,
        help="Hide missing metadata warnings and errors to stdout/stderr",
    )
    parser.add_argument(
        "--show-examples",
        "-E",
        "--HELP",
        "-H",
        help="Shows example usage of this cli.",
        action="store_true",
    )

    parser.add_argument(
        "--set-dcm2niix-path",
        help="Provide a path to a dcm2niix install/exe, writes path to config "
        f"file {Path.home()}/.pet2bidsconfig under the variable "
        f"DCM2NIIX_PATH",
        type=pathlib.Path,
    )
    parser.add_argument(
        "--set-dcm2niix-options",
        help="Provide dcm2niix options to be used as defaults, writes to config "
        f"file {Path.home()}/.pet2bidsconfig under the variable "
        f"PET2BIDS_DCM2NIIX_OPTIONS. Example: --set-dcm2niix-options '-v y -w 1 -z y'",
        type=str,
    )
    parser.add_argument(
        "--set-default-metadata-json",
        help="Provide a path to a default metadata file json file."
        "This file will be used to fill in missing metadata not"
        "contained within dicom headers or spreadsheet metadata."
        "Sets given path to DEFAULT_METADATA_JSON var in "
        f"{Path.home()}/.pet2bidsconfig",
    )
    parser.add_argument(
        "--trc",
        "--tracer",
        type=str,
        default="",
        help="Provide a tracer name to be used in the output file name",
    )
    parser.add_argument(
        "--run",
        type=str,
        default="",
        help="Provide a run id to be used in the output file name",
    )
    parser.add_argument(
        "--rec",
        type=str,
        default="",
        help="Provide a reconstruction method to be used in the output file name",
    )
    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=f"{helper_functions.get_version()}",
    )
    parser.add_argument(
        "--notrack",
        action="store_true",
        default=False,
        help="Opt-out of sending tracking information of this run to the PET2BIDS developers. "
        "This information helps to improve PET2BIDS and provides an indicator of real world "
        "usage crucial for obtaining funding.",
    )
    parser.add_argument(
        "--ezbids",
        action="store_true",
        default=False,
        help="Add fields to json output that are useful for ezBIDS or other conversion software. This will de-anonymize"
        " pet2bids output and add AcquisitionDate an AcquisitionTime into the output json.",
    )
    parser.add_argument(
        "--ignore-dcm2niix-errors",
        action="store_true",
        default=False,
        help="Accept any NifTi produced by dcm2niix even if it contains errors. This flag should only be used for "
        "batch processing and only if you're performing robust QC after the fact.",
    )
    parser.add_argument(
        "--dcm2niix-options",
        nargs="*",
        default=[],
        help="Additional dcm2niix options to pass through. Use the same format as dcm2niix command line. "
        "Note: -f flag and its value will be filtered out as file format is handled separately. "
        "You can also set default options using --set-dcm2niix-options or the PET2BIDS_DCM2NIIX_OPTIONS environment variable. "
        "Example: --dcm2niix-options -v y -w 1 -z y",
    )
    return parser


example1 = textwrap.dedent(
    """

Usage examples are below, the first being the most brutish way of making dcm2niix4pet to pass through the
BIDS validator (with no errors, removing all warnings is left to the user as an exercise) see:

example 1 (Passing PET metadata via the --kwargs argument): 
    
    # Note `#` denotes a comment
    # dcm2niix4pet is called with the following arguments
    
    # folder -> GeneralElectricSignaPETMR-NIMH/
    # destination-path -> sub-GeneralElectricSignaPETMRINIMH/pet
    # kwargs -> a bunch of key pair arguments spaced 1 space apart with the values surrounded by double quotes

    dcm2niix4pet GeneralElectricSignaPETMR-NIMH/ --destination-path sub-GeneralElectricSignaPETMRNIMH/pet 
    --kwargs TimeZero="14:08:45" Manufacturer="GE MEDICAL SYSTEMS" ManufacturersModelName="SIGNA PET/MR" 
    InstitutionName="NIH Clinical Center, USA" BodyPart="Phantom" Units="Bq/mL" TracerName="Gallium citrate" 
    TracerRadionuclide="Germanium68" InjectedRadioactivity=1 SpecificRadioactivity=23423.75 
    ModeOfAdministration="infusion" FrameTimesStart=0 
    AcquisitionMode="list mode" ImageDecayCorrected="False" FrameTimesStart="[0]" ImageDecayCorrectionTime=0 
    ReconMethodParameterValues="[1, 1]" ReconFilterType="n/a" ReconFilterSize=1

    # The output of the above command (given some GE phantoms from the NIMH) can be seen below with tree
    
    tree sub-GeneralElectricSignaPETMRNIMH 
    sub-GeneralElectricSignaPETMRNIMH
    └── pet
        ├── sub-GeneralElectricSignaPETMRNIMH_pet.json
        └── sub-GeneralElectricSignaPETMRNIMH_pet.nii.gz

    # Further, when we examine the json output file we can see that all of our metadata supplied via kwargs was written
    # into the sidecar json 
    
    cat sub-GeneralElectricSignalPETMRNIMH/pet/sub-GeneralElectricSignaPETMRNIMH_pet.json
    {
        "Modality": "PT",
        "Manufacturer": "GE MEDICAL SYSTEMS",
        "ManufacturersModelName": "SIGNA PET/MR",
        "InstitutionName": "NIH Clinical Center, USA",
        "StationName": "PETMR",
        "PatientPosition": "HFS",
        "SoftwareVersions": "61.00",
        "SeriesDescription": "PET Scan for VQC Verification",
        "ProtocolName": "PET Scan for VQC Verification",
        "ImageType": [
            "ORIGINAL",
            "PRIMARY"
        ],
        "SeriesNumber": 2,
        "Radiopharmaceutical": "Germanium",
        "RadionuclidePositronFraction": 0.891,
        "RadionuclideHalfLife": 23410100.0,
        "Units": "Bq/mL",
        "DecayCorrection": "NONE",
        "AttenuationCorrectionMethod": "NONE, 0.000000 cm-1,",
        "SliceThickness": 2.78,
        "ImageOrientationPatientDICOM": [
            1,
            0,
            0,
            0,
            1,
            0
        ],
        "ConversionSoftware": [
            "dcm2niix",
            "pypet2bids"
        ],
        "ConversionSoftwareVersion": [
            "v1.0.20220720",
            "1.0.2"
        ],
        "FrameDuration": [
            98000
        ],
        "ReconMethodName": "VPFX",
        "ReconMethodParameterUnits": [
            "none",
            "none"
        ],
        "ReconMethodParameterLabels": [
            "subsets",
            "iterations"
        ],
        "ReconMethodParameterValues": [
            1,
            1
        ],
        "AttenuationCorrection": "NONE, 0.000000 cm-1,",
        "TimeZero": "14:08:45",
        "ScanStart": 0,
        "InjectionStart": 0,
        "TracerRadionuclide": "Germanium68",
        "BodyPart": "Phantom",
        "TracerName": "Gallium citrate",
        "InjectedRadioactivity": 1,
        "SpecificRadioactivity": 23423.75,
        "ModeOfAdministration": "infusion",
        "FrameTimesStart": [
            0
        ],
        "AcquisitionMode": "list mode",
        "ImageDecayCorrected": false,
        "ImageDecayCorrectionTime": 0,
        "ReconFilterType": "n/a",
        "ReconFilterSize": 1,
        "InjectedRadioactivityUnits": "MBq",
        "SpecificRadioactivityUnits": "Bq/g",
        "InjectedMass": "n/a",
        "InjectedMassUnits": "n/a"
        }"""
)


def main():
    """
    Executes cli() and uses Dcm2niix4PET class to convert a folder containing dicoms into nifti + json.

    :return: None
    """

    # collect args
    cli_parser = cli()

    if len(sys.argv) == 1:
        cli_parser.print_usage()
        print(f"version: {helper_functions.get_version()}")
        sys.exit(1)
    else:
        cli_args = cli_parser.parse_args()

    if cli_args.silent:
        logger.disabled = True

    if cli_args.show_examples:
        print(example1)
        sys.exit(0)

    if cli_args.set_dcm2niix_path:
        helper_functions.modify_config_file("DCM2NIIX_PATH", cli_args.set_dcm2niix_path)
        sys.exit(0)

    if cli_args.set_dcm2niix_options:
        helper_functions.modify_config_file(
            "PET2BIDS_DCM2NIIX_OPTIONS", cli_args.set_dcm2niix_options
        )
        sys.exit(0)

    if cli_args.set_default_metadata_json:
        helper_functions.modify_config_file(
            "DEFAULT_METADATA_JSON", cli_args.set_default_metadata_json
        )
        sys.exit(0)
    if cli_args.notrack:
        environ["PET2BIDS_TRACK"] = "False"

    elif cli_args.folder:
        # instantiate class
        converter = Dcm2niix4PET(
            image_folder=helper_functions.expand_path(cli_args.folder),
            destination_path=helper_functions.expand_path(cli_args.destination_path),
            metadata_path=helper_functions.expand_path(cli_args.metadata_path),
            metadata_translation_script=helper_functions.expand_path(
                cli_args.translation_script_path
            ),
            additional_arguments=cli_args.kwargs,
            dcm2niix_options=(
                " ".join(cli_args.dcm2niix_options) if cli_args.dcm2niix_options else ""
            ),
            tempdir_location=cli_args.tempdir,
            silent=cli_args.silent,
            ezbids=cli_args.ezbids,
            ignore_dcm2niix_errors=cli_args.ignore_dcm2niix_errors,
        )

        if cli_args.trc:
            converter.tracer = "trc-" + cli_args.trc
        if cli_args.run:
            converter.run_id = "run-" + cli_args.run
        if cli_args.rec:
            converter.reconstruction_method = "rec-" + cli_args.rec

        converter.convert()
    else:
        print(
            "folder is a required argument for running dcm2niix, see -h for more detailed usage."
        )
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
This program will create an ecat file if provided an ecat schema and a dictionary of values to populate that schema
with.

First this program collects the same schemas that read_ecat.py does, from read_ecat import ecat_header_maps.
Next this program selects one of the header maps as specified by some input e.g. if given: ecat7.3 it would 
select the standard image matrix at: ecat_header_maps['ecat_headers']['73']['mainheader'] 
and the subheader map at: ecat_header_maps['ecat_headers']['73']['11'] or whatever number is corresponding to 
the type of ecat header you wish to write. Perhaps these should belong in a reverse sort of dictionary going to
need to create the directory byte block(s) for an ecat. Basically, reverse the process of reading the 
directory in lines 227 through 257 in ecat_read. 
    
    determine the number of frames in the image/pixel data
    create empty table(s) of dtype >i4 dimensions of 4 rows by 64 columns 
    fill empty tables w/ zeros
    place the number 2 in the second row, first column of the table if it's the only directory table or the last 
    directory
    table.
    final output should be a bytes: 1024 type object.
    
After generating the table you should then be able to write the main header, write the directory table,
then write each subheader and corresponding pixel data


:Author: Anthony Galassi

:Copyright: Open NeuroPET team
"""

import struct
from math import ceil, floor
from pypet2bids.read_ecat import ecat_header_maps, get_buffer_size
import numpy
from pathlib import Path


def write_header(
    ecat_file,
    schema: dict,
    values: dict = {},
    byte_position: int = 0,
    seek: bool = False,
):
    """
    Given a filepath and a schema this function will write an ecat header to a file. If supplied a dictionary of values
    it will populate the written header with them, if not it will fill the header with empty values. If a byte position
    argument is supplied write_header will write to the byte position in the file supplied at file path, else it will
    seek to the end of the file and write there.
    :param ecat_file: the ecat file to be written
    :param schema: dictionary schema of the ecat header
    :param values: dictionary of values corresponding to the 'variable_name' entries in the schema dictionary
    :param byte_position: offset to write the header at, default is at the zero'th byte position.
    :param seek: if True use the provided byte position argument else, collect byte position form opened ecat_file
    :return: the end byte position, spoilers this will be 512 bytes offset from byte_position if that argument is
    supplied.
    """

    if not seek:
        byte_position = ecat_file.tell()

    for entry in schema:
        byte_position, variable_name, struct_fmt = (
            entry["byte"],
            entry["variable_name"],
            entry["struct"],
        )
        struct_fmt = ">" + struct_fmt
        byte_width = struct.calcsize(struct_fmt)
        value_to_write = values.get(variable_name, None)

        # if no value is supplied in the value dict, pack with empty bytes as well
        if value_to_write is None:
            pack_empty = True
        # set variable to false if neither of these conditions is met
        else:
            pack_empty = False

        # for fill or empty values supplied in the values dictionary
        if pack_empty:
            fill = byte_width * "x"
            ecat_file.write(struct.pack(fill))
        elif "s" in struct_fmt:
            value_to_write = bytes(value_to_write, "ascii")
            ecat_file.write(struct.pack(struct_fmt, value_to_write))
        # for all other cases
        else:
            if type(value_to_write) is tuple and len(value_to_write) > 1:
                value_to_write = list(value_to_write)
            elif type(value_to_write) is not list:
                value_to_write = [value_to_write]
            try:
                ecat_file.write(struct.pack(struct_fmt, *value_to_write))
            except struct.error as err:
                if values["DATA_TYPE"] == 5 and "MAX" in variable_name:
                    value_to_write = 32767
                elif values["DATA_TYPE"] == 5 and "MIN" in variable_name:
                    value_to_write = -32767
                    print(
                        "Uncertain how to handle min and max datatypes for float arrays when writing ecats.\n"
                        "if you know more about what header min and max values should be in the case of an float32\n"
                        "image matrix please consider making a pull request to this library or posting an issue."
                    )
                else:
                    print(
                        f"Oh no {value_to_write} is out of range for {struct_fmt}, variable {variable_name}"
                    )
                    raise err
    return byte_width + byte_position


def create_directory_table(
    num_frames: int = 0, pixel_dimensions: dict = {}, pixel_byte_size: int = 2
):
    """
    Creates directory tables for an ecat file when provided with the number of frames, the total number of pixels
    per frame, and the byte size of each of those pixels. Ecat's can have int16 (2 byte widths) or float 32 (4 byte
    widths) pixel values. This table will then be used to look up each subheader and it's corresponding pixel data
    after it is written.
    :param num_frames: number of frames taken during the ecat scan, a.k.a. time dimension
    :param pixel_dimensions: a dictionary such as {'x': 2, 'y': 2, 'z': 2}, the relevant information is contained within
    the values, not the keys. However, those keys may be important w/ cylindrical ecat files, at this point it is
    undetermined.
    :param pixel_byte_size: the width of each pixel's value in bytes.
    :return: a list containing 2 dimensional directory tables populated w/ elements corresponding to the byte position
    of the frame info of the ECAT.
    """
    # first determine the number of byte blocks the directory table(s) require based on the
    # number of frames
    required_directory_blocks = ceil(num_frames / 31)

    # determine the width of the frame byte blocks with the pixel dimensions and data sizes
    pixel_volume = numpy.product([*pixel_dimensions.values()])

    # still not sure what some of these numbers mean, but we populate the first column of the byte array with them
    directory_tables = []

    # create entries for the directory table
    directory_order = [i + 1 for i in range(num_frames)]

    # why is the initial table byte position equal to 3? Because the first header and pixel data lies
    # after the main header byte block at position 0 to 1 and the directory table itself occupies byte blocks
    # 1 to 2, thus the first frame (header and pixel data) will land at byte block number 3! Whoopee!
    table_byte_position = 3  # 1 + required_directory_blocks
    for i in range(required_directory_blocks):
        # initialize empty 4 x 32 array
        table = numpy.ndarray((4, 32), dtype=">i4")

        # populate first column of table with codes and undetermined codes
        # note these table values are only extrapolated from a data set w/ 45 frames and datasets with less than
        # 31 frames. Behavior or values for frames numbering above 45 (or more specifically 62) is unknown.
        if i == (required_directory_blocks - 1):
            frames_to_iterate = num_frames % 31
            table[0, 0] = (
                31 - frames_to_iterate
            )  # number of un-used columns in the directory table
            table[1, 0] = 2  # stop value
            table[2, 0] = 2  # stop value
            table[3, 0] = frames_to_iterate  # number of frames referenced in this table
        else:
            frames_to_iterate = floor(num_frames / 31) * 31
            table[0, 0] = 0
            table[1, 0] = (
                0  # note this is just a placeholder the real value is entered after the final position of the
            )
            # last entry in the first directory is calculated
            table[2, 0] = 0
            table[3, 0] = frames_to_iterate

        for column in range(frames_to_iterate):
            if column != 0:
                table_byte_position += 1
            elif column == 0 and i == 1:
                table_byte_position += 2
            table[0, column + 1] = directory_order.pop(0)
            table[1, column + 1] = table_byte_position
            # frame byte position is shifted
            table_byte_position = int(
                (pixel_byte_size * pixel_volume) / 512 + table_byte_position
            )
            table[2, column + 1] = table_byte_position
            table[3, column + 1] = 1

        # VERY IMPORTANT
        if i != (required_directory_blocks - 1):
            table[1, 0] = table[2, column + 1] + 1

        directory_tables.append(table)

    return directory_tables


def write_directory_table(file, directory_tables: list, seek: bool = False):
    """
    Given a list of numpy.ndarray tables and a file path, this function flattens the directory tables to be written
    into the filepath as bytes strings, then writes the bytes strings to the end of the file, returning the byte position
    it left off on.
    :param file: an ecat file that has been initialized with it's main header information
    :param directory_tables: a list of directory tables (1 to 2) that are 4x32 in dimensions and of dtype int32.
    :param seek: seek to directory table position at 512 bytes, defaults to writing at current position of
    file read head.
    :return: The byte position after writing
    """
    # flatten the lists into byte strings

    for n, table in enumerate(directory_tables):
        # first table lies directly after the main header byte block, which is 512 bytes from the start of the file

        if n == 0:
            if seek:
                file.seek(512)
            else:
                pass
        # additional directory table positions are recorded in the previous directory table in row 1 column 0
        else:
            next_table_position = directory_tables[n - 1][1, 0] - 1
            file.seek(512 * next_table_position)
        transpose_table = numpy.transpose(table)
        flattened_transpose = numpy.reshape(transpose_table, (4, -1)).flatten()

        for index in range(flattened_transpose.size):
            file.write(struct.pack(">l", flattened_transpose[index]))

    return file.tell()


def write_pixel_data(
    ecat_file, pixel_data: numpy.ndarray, byte_position: int = None, seek: bool = None
):
    if seek and byte_position:
        ecat_file.seek(byte_position)
    # elif (seek and not byte_position) or (byte_position and not seek):
    # raise Exception("Must provide seek boolean and byte position")
    else:
        pass
    flattened_pixels = pixel_data.flatten().tobytes()
    ecat_file.write(flattened_pixels)
    return 0


def write_ecat(
    ecat_file: Path,
    mainheader_schema: dict,
    mainheader_values: dict,
    subheaders_values: list,
    subheader_schema: dict,
    number_of_frames: int,
    pixel_x_dimension: int,
    pixel_y_dimension: int,
    pixel_z_dimension: int,
    pixel_byte_size: int,
    pixel_data: list = [],
):

    # open the ecat file!
    with open(ecat_file, "w+b") as outfile:

        # first things first, write the main header with supplied information
        write_header(
            ecat_file=outfile, schema=mainheader_schema, values=mainheader_values
        )
        position_post_header_write = outfile.tell()
        # create the directory table
        directory_table = create_directory_table(
            num_frames=number_of_frames,
            pixel_dimensions={
                "x": pixel_x_dimension,
                "y": pixel_y_dimension,
                "z": pixel_z_dimension,
            },
            pixel_byte_size=pixel_byte_size,
        )
        # write the directory table to the file
        write_directory_table(file=outfile, directory_tables=directory_table)

        position_post_table_write = outfile.tell()
        # write subheaders followed by pixel data
        for index, subheader in enumerate(subheaders_values):
            position = outfile.tell()
            table_position = directory_table[0][1, index + 1] * 512
            write_header(
                ecat_file=outfile,
                schema=subheader_schema,
                values=subheader,
                byte_position=outfile.tell(),
            )
            position_post_subheader_write = outfile.tell()
            write_pixel_data(ecat_file=outfile, pixel_data=pixel_data[index])
            position_post_subheader_pixel_data_write = outfile.tell()

    return ecat_file



from pathlib import Path
from os.path import join
from os import listdir
import json
from json_maj.main import JsonMAJ, load_json_or_dict
import re
from dateutil import parser
import argparse
import pydicom
import datetime
from typing import Union
from pandas import Timestamp

try:
    import helper_functions
    import is_pet
    import pet_metadata as metadata
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions
    import pypet2bids.is_pet as is_pet
    import pypet2bids.pet_metadata as metadata

# import logging
logger = helper_functions.logger("pypet2bids")

# create a dictionary to house the PET metadata files
metadata_dictionaries = {
    "blood_metadata": metadata.blood_metadata,
    "dicom2bids": metadata.dicom2bids,
    "PET_reconstruction_methods": metadata.PET_reconstruction_methods,
    "schema": metadata.schema,
    "PET_metadata": metadata.PET_metadata,
}


def check_json(
    path_to_json,
    items_to_check=None,
    silent=False,
    spreadsheet_metadata={},
    mandatory=True,
    recommended=True,
    logger_name="pypet2bids",
    **additional_arguments,
):
    """
    This method opens a json and checks to see if a set of mandatory values is present within that json, optionally it
    also checks for recommended key value pairs. If fields are not present a warning is raised to the user.

    :param spreadsheet_metadata:
    :type spreadsheet_metadata:
    :param path_to_json: path to a json file e.g. a BIDS sidecar file created after running dcm2niix
    :param items_to_check: a dictionary with items to check for within that json. If None is supplied defaults to the
           PET_metadata imported from pet_metadata.PET_metadata
    :param silent: Raises warnings or errors to stdout if this flag is set to True
    :return: dictionary of items existence and value state, if key is True/False there exists/(does not exist) a
            corresponding entry in the json the same can be said of value
    """

    logger = helper_functions.logger(logger_name)

    if silent:
        logger.disabled = True
    else:
        logger.disabled = False

    # check if path exists
    path_to_json = Path(path_to_json)
    if not path_to_json.exists():
        raise FileNotFoundError(path_to_json)

    # check for default argument for dictionary of items to check
    if items_to_check is None:
        items_to_check = metadata_dictionaries["PET_metadata"]
        # remove blood tsv data from items to check
        if items_to_check.get("blood_recording_fields", None):
            items_to_check.pop("blood_recording_fields")

    # open the json
    with open(path_to_json, "r") as in_file:
        json_to_check = json.load(in_file)

    # initialize warning colors and warning storage dictionary
    storage = {}
    flattened_spreadsheet_metadata = {}
    flattened_spreadsheet_metadata.update(spreadsheet_metadata.get("nifti_json", {}))
    flattened_spreadsheet_metadata.update(spreadsheet_metadata.get("blood_json", {}))
    flattened_spreadsheet_metadata.update(spreadsheet_metadata.get("blood_tsv", {}))

    if mandatory:
        for item in items_to_check["mandatory"]:
            all_good = False
            if (
                item in json_to_check.keys()
                and (
                    json_to_check.get(item, None) is not None
                    or json_to_check.get(item) != ""
                )
                or item in additional_arguments
                or item in flattened_spreadsheet_metadata.keys()
            ):
                # this json has both the key and a non-blank value do nothing
                all_good = True
                pass
            elif item in json_to_check.keys() and (
                json_to_check.get(item, None) is None
                or json_to_check.get(item, None) == ""
            ):
                logger.error(f"{item} present but has null value.")
                storage[item] = {"key": True, "value": False}
            elif not all_good:
                logger.error(
                    f"{item} is not present in {path_to_json}. This will have to be "
                    f"corrected post conversion."
                )
                storage[item] = {"key": False, "value": False}

    if recommended:
        for item in items_to_check["recommended"]:
            all_good = False
            if (
                item in json_to_check.keys()
                and (
                    json_to_check.get(item, None) is not None
                    or json_to_check.get(item) != ""
                )
                or item in additional_arguments
                or item in flattened_spreadsheet_metadata.keys()
            ):
                # this json has both the key and a non-blank value do nothing
                all_good = True
                pass
            elif item in json_to_check.keys() and (
                json_to_check.get(item, None) is None
                or json_to_check.get(item, None) == ""
            ):
                logger.info(f"{item} present but has null value.")
                storage[item] = {"key": True, "value": False}
            elif not all_good:
                logger.info(f"{item} is recommended but not present in {path_to_json}")
                storage[item] = {"key": False, "value": False}

    return storage


def update_json_with_dicom_value(
    path_to_json,
    missing_values,
    dicom_header,
    dicom2bids_json=None,
    silent=True,
    ezbids=False,
    **additional_arguments,
):
    """
    We go through all the missing values or keys that we find in the sidecar json and attempt to extract those
    missing entities from the dicom source. This function relies on many heuristics a.k.a. many unique conditionals and
    simply is what it is, hate the game not the player.

    :param path_to_json: path to the sidecar json to check
    :param missing_values: dictionary output from check_json indicating missing fields and/or values
    :param dicom_header: the dicom or dicoms that may contain information not picked up by dcm2niix
    :param dicom2bids_json: a json file that maps dicom header entities to their corresponding BIDS entities
    :param silent: run silently without error, status, or warning messages
    :param ezbids: boolean to supply additional data that ezbids or other software requires, defaults to false. When
    true the sidecar json will be updated with AcquisitionDate, AcquisitionTime, and AcquisitionDateTime
    :return: a dictionary of successfully updated (written to the json file) fields and values
    """

    if silent:
        logger.disabled = True

    json_sidecar_path = Path(path_to_json)
    if not json_sidecar_path.exists():
        with open(path_to_json, "w") as outfile:
            json.dump({}, outfile)

    # load the sidecar json
    sidecar_json = load_json_or_dict(str(path_to_json))

    # purely to clean up the generated read the docs page from sphinx, otherwise the entire json appears in the
    # read the docs page.
    if dicom2bids_json is None:
        dicom2bids_json = metadata_dictionaries["dicom2bids"]

    # Units gets written as Unit in older versions of dcm2niix here we check for missing Units and present Unit entity
    units = missing_values.get("Units", None)
    if units:
        try:
            # Units is missing, check to see if Unit is present
            if sidecar_json.get("Unit", None):
                temp = JsonMAJ(
                    path_to_json, {"Units": sidecar_json.get("Unit")}, bids_null=True
                )
                temp.remove("Unit")
            else:  # we source the Units value from the dicom header and update the json
                JsonMAJ(path_to_json, {"Units": dicom_header.Units}, bids_null=True)
        except AttributeError:
            logger.error(
                f"Dicom is missing Unit(s) field, are you sure this is a PET dicom?"
            )
    # pair up dicom fields with bids sidecar json field, we do this in a separate json file
    # it's loaded when this script is run and stored in metadata dictionaries
    dcmfields = dicom2bids_json["dcmfields"]
    jsonfields = dicom2bids_json["jsonfields"]

    regex_cases = ["ReconstructionMethod", "ConvolutionKernel"]

    # strip excess characters from dcmfields
    dcmfields = [re.sub("[^0-9a-zA-Z]+", "", field) for field in dcmfields]
    paired_fields = {}
    for index, field in enumerate(jsonfields):
        paired_fields[field] = dcmfields[index]

    logger.info("Attempting to locate missing BIDS fields in dicom header")
    # go through missing fields and reach into dicom to pull out values
    json_updater = JsonMAJ(json_path=path_to_json, bids_null=True)
    for key, value in paired_fields.items():
        missing_bids_field = missing_values.get(key, None)
        # if field is missing look into dicom
        if missing_bids_field and key not in additional_arguments:
            # there are a few special cases that require regex splitting of the dicom entries
            # into several bids sidecar entities
            try:
                dicom_field = getattr(dicom_header, value)
                logger.info(f"FOUND {value} corresponding to BIDS {key}: {dicom_field}")
            except AttributeError:
                dicom_field = None
                logger.info(
                    f"NOT FOUND {value} corresponding to BIDS {key} in dicom header."
                )

            if dicom_field and value in regex_cases:
                # if it exists get rid of it, we don't want no part of it.
                if sidecar_json.get("ReconMethodName", None):
                    json_updater.remove("ReconstructionMethod")
                if dicom_header.get("ReconstructionMethod", None):
                    reconstruction_method = dicom_header.ReconstructionMethod
                    json_updater.remove("ReconstructionMethod")
                    reconstruction_method = helper_functions.get_recon_method(
                        reconstruction_method
                    )

                    json_updater.update(reconstruction_method)

            elif dicom_field:
                # update json
                json_updater.update({key: dicom_field})

    # Additional Heuristics are included below

    # See if time zero is missing in json or additional args
    if missing_values.get("TimeZero", None):
        if (
            missing_values.get("TimeZero")["key"] is False
            or missing_values.get("TimeZero")["value"] is False
        ):
            time_parser = parser
            if sidecar_json.get("AcquisitionTime", None):
                acquisition_time = (
                    time_parser.parse(sidecar_json.get("AcquisitionTime"))
                    .time()
                    .strftime("%H:%M:%S")
                )
            else:
                acquisition_time = (
                    time_parser.parse(dicom_header["SeriesTime"].value)
                    .time()
                    .strftime("%H:%M:%S")
                )

            json_updater.update({"TimeZero": acquisition_time})
            json_updater.remove("AcquisitionTime")
            json_updater.update({"ScanStart": 0})
        else:
            pass

    if missing_values.get("ScanStart", None):
        if (
            missing_values.get("ScanStart")["key"] is False
            or missing_values.get("ScanStart")["value"] is False
        ):
            json_updater.update({"ScanStart": 0})
    if missing_values.get("InjectionStart", None):
        if (
            missing_values.get("InjectionStart")["key"] is False
            or missing_values.get("InjectionStart")["value"] is False
        ):
            json_updater.update({"InjectionStart": 0})

    # check to see if units are BQML
    json_updater = JsonMAJ(str(path_to_json), bids_null=True)
    if json_updater.get("Units") == "BQML":
        json_updater.update({"Units": "Bq/mL"})

    # Add radionuclide to json
    Radionuclide = get_radionuclide(dicom_header)
    if Radionuclide:
        json_updater.update({"TracerRadionuclide": Radionuclide})

    # remove scandate if it exists
    json_updater.remove("ScanDate")

    # lastly if ezbids is true update the sidecar with acquisition data
    if ezbids:
        acquisition_date = parser.parse(dicom_header.get("AcquisitionDate", ""))
        acquisition_time = parser.parse(dicom_header.get("AcquisitionTime", ""))
        if acquisition_time and acquisition_date:
            acquisition_datetime = datetime.datetime.combine(
                acquisition_date.date(), acquisition_time.time()
            )
        else:
            acquisition_datetime = "0000-00-00T00:00:00"
        json_updater.update(
            {
                "AcquisitionDate": f"{acquisition_date.date()}",
                "AcquisitionTime": f"{acquisition_time.time()}",
                "AcquisitionDateTime": f"{acquisition_datetime.isoformat()}",
            }
        )

    # after updating raise warnings to user if values in json don't match values in dicom headers, only warn!
    updated_values = json.load(open(path_to_json, "r"))
    for key, value in paired_fields.items():
        try:
            json_field = updated_values.get(key)
            dicom_field = dicom_header.__getattr__(key)
            if json_field != dicom_field:
                logger.info(
                    f"WARNING!!!! JSON Field {key} with value {json_field} does not match dicom value "
                    f"of {dicom_field}"
                )
        except AttributeError:
            pass


def update_json_with_dicom_value_cli():
    """
    Command line interface for update_json_with_dicom_value, updates a PET json with values from a dicom header,
    optionally can update with values from a spreadsheet or via values passed in as additional arguments with the -k
    --additional_arguments flag. This command be accessed after installation of pypet2bids via
    `updatepetjsonfromdicom`.
    """
    dicom_update_parser = argparse.ArgumentParser(
        description="Updates a PET json with values from a dicom header "
        "optionally can update with values from a spreadsheet or "
        "via values passed in as additional arguments"
    )
    dicom_update_parser.add_argument(
        "-j", "--json", help="path to json to update", required=True
    )
    dicom_update_parser.add_argument(
        "-d", "--dicom", help="path to dicom to extract values from", required=True
    )
    dicom_update_parser.add_argument(
        "-m", "--metadata-path", help="path to metadata json", default=None
    )
    dicom_update_parser.add_argument(
        "-k",
        "--additional_arguments",
        help="additional key value pairs to update json with",
        nargs="*",
        action=helper_functions.ParseKwargs,
        default={},
    )

    args = dicom_update_parser.parse_args()

    try:
        # get missing values
        missing_values = check_json(args.json, silent=True)
    except FileNotFoundError:
        with open(args.json, "w") as outfile:
            json.dump({}, outfile)
        missing_values = check_json(args.json, silent=True)

    # load dicom header
    dicom_header = pydicom.dcmread(args.dicom, stop_before_pixels=True)

    if args.metadata_path:
        spreadsheet_metadata = get_metadata_from_spreadsheet(
            args.metadata_path, args.dicom, dicom_header, **args.additional_arguments
        )
        # update json
        update_json_with_dicom_value(
            args.json,
            missing_values,
            dicom_header,
            silent=True,
            spreadsheet_metadata=spreadsheet_metadata["nifti_json"],
            **args.additional_arguments,
        )

        JsonMAJ(args.json, update_values=spreadsheet_metadata["nifti_json"]).update()
    else:
        update_json_with_dicom_value(
            args.json,
            missing_values,
            dicom_header,
            silent=True,
            **args.additional_arguments,
        )

    JsonMAJ(args.json, update_values=args.additional_arguments).update()

    # check json again after updating
    check_json(
        args.json,
        required=True,
        recommended=True,
        silent=False,
        logger_name="check_json",
    )


def update_json_cli():
    """
    Updates a json file with user supplied values or values from a spreadsheet. This command can be accessed after
    conversion via `updatepetjson` if so required.
    """
    update_json_parser = argparse.ArgumentParser(
        description="Updates a json file with user supplied values or values "
        "from a spreadsheet. This command can be accessed after "
        "conversion via `updatepetjson` if so required. Primarily"
        "intended to 'touch up' or create sidecar json files."
    )
    update_json_parser.add_argument(
        "-j", "--json", help="path to json to update", required=True
    )
    update_json_parser.add_argument(
        "-k",
        "--additional_arguments",
        help="additional key value pairs to update json with",
        nargs="*",
        action=helper_functions.ParseKwargs,
        default={},
    )
    update_json_parser.add_argument(
        "-m", "--metadata-path", help="path to metadata json", default=None
    )

    update_json_args = update_json_parser.parse_args()

    if update_json_args.metadata_path:
        nifti_sidecar_metadata = (
            get_metadata_from_spreadsheet(
                update_json_args.metadata_path, update_json_args.json
            )
        )["nifti_json"]
    else:
        nifti_sidecar_metadata = {}

    j = JsonMAJ(update_json_args.json, update_values=nifti_sidecar_metadata)
    j.update()
    j.update(update_json_args.additional_arguments)

    # check meta radio inputs
    j.update(check_meta_radio_inputs(j.json_data))

    # check json again after updating
    check_json(
        update_json_args.json,
        required=True,
        recommended=True,
        silent=False,
        logger_name="check_json",
    )


def get_radionuclide(pydicom_dicom):
    """
    Gets the radionuclide if given a pydicom_object if
    pydicom_object.RadiopharmaceuticalInformationSequence[0].RadionuclideCodeSequence exists

    :param pydicom_dicom: dicom object collected by pydicom.dcmread("dicom_file.img")
    :return: Labeled Radionuclide e.g. 11Carbon, 18Flourine
    """
    radionuclide = ""
    try:
        radiopharmaceutical_information_sequence = (
            pydicom_dicom.RadiopharmaceuticalInformationSequence
        )
        radionuclide_code_sequence = radiopharmaceutical_information_sequence[
            0
        ].RadionuclideCodeSequence
        code_value = radionuclide_code_sequence[0].CodeValue
        code_meaning = radionuclide_code_sequence[0].CodeMeaning
        extraction_good = True
    except AttributeError:
        logger.info(
            "Unable to extract RadioNuclideCodeSequence from RadiopharmaceuticalInformationSequence"
        )
        extraction_good = False

    if extraction_good:
        # check to see if these nucleotides appear in our verified values
        verified_nucleotides = metadata_dictionaries["dicom2bids"]["RadionuclideCodes"]

        check_code_value = ""
        check_code_meaning = ""

        if code_value in verified_nucleotides.keys():
            check_code_value = code_value
        else:
            logger.warning(
                f"Radionuclide Code {code_value} does not match any known codes in dcm2bids.json\n"
                f"will attempt to infer from code meaning {code_meaning}"
            )

        if code_meaning in verified_nucleotides.values():
            radionuclide = re.sub(r"\^", "", code_meaning)
            check_code_meaning = code_meaning
        else:
            logger.warning(
                f"Radionuclide Meaning {code_meaning} not in known values in dcm2bids json"
            )
            if code_value in verified_nucleotides.keys():
                radionuclide = re.sub(r"\^", "", verified_nucleotides[code_value])

        # final check
        if check_code_meaning and check_code_value:
            pass
        else:
            logger.warning(
                f"WARNING!!!! Possible mismatch between nuclide code meaning {code_meaning} and {code_value} in dicom "
                f"header"
            )

    return radionuclide


def check_meta_radio_inputs(kwargs: dict, logger="pypet2bids") -> dict:
    """
    Routine to check input consistency, possibly generate new ones from PET
    BIDS metadata - this only makes sense if you respect the input units as
    indicated

    e.g. arguments in are provided via the following params (key/value pairs)
        - 'InjectedRadioctivity',81.24
        - 'SpecificRadioactivity',1.3019e+04

    Units are transformed as follows:

    InjectedRadioactivity: in MBq
    InjectedMass:          in ug
    SpecificRadioactivity: in Bq/g or MBq/ug
    MolarActivity:         in GBq/umol
    MolecularWeight:       in g/mol

    :param kwargs: metadata key pair's to examine
    :type kwargs: dict
    :return: fitted/massaged metadata, return type is an update on input `kwargs`
    :rtype: dict
    """

    logger = helper_functions.logger(logger)

    InjectedRadioactivity = kwargs.get("InjectedRadioactivity", None)
    InjectedMass = kwargs.get("InjectedMass", None)
    SpecificRadioactivity = kwargs.get("SpecificRadioactivity", None)
    MolarActivity = kwargs.get("MolarActivity", None)
    MolecularWeight = kwargs.get("MolecularWeight", None)

    data_out = {}

    if InjectedRadioactivity and InjectedMass:
        data_out["InjectedRadioactivity"] = InjectedRadioactivity
        data_out["InjectedRadioactivityUnits"] = "MBq"
        data_out["InjectedMass"] = InjectedMass
        data_out["InjectedMassUnits"] = "ug"
        # check for strings where there shouldn't be strings
        numeric_check = [
            helper_functions.is_numeric(str(InjectedRadioactivity)),
            helper_functions.is_numeric(str(InjectedMass)),
        ]
        if False in numeric_check:
            data_out["InjectedMass"] = "n/a"
            data_out["InjectedMassUnits"] = "n/a"
        else:
            tmp = (InjectedRadioactivity * 10**6) / (InjectedMass * 10**6)
            if SpecificRadioactivity:
                if SpecificRadioactivity != tmp:
                    logger.warning(
                        "Inferred SpecificRadioactivity in Bq/g doesn't match InjectedRadioactivity "
                        "and InjectedMass, could be a unit issue"
                    )
                data_out["SpecificRadioactivity"] = SpecificRadioactivity
                data_out["SpecificRadioactivityUnits"] = kwargs.get(
                    "SpecificRadioactivityUnityUnits", "n/a"
                )
            else:
                data_out["SpecificRadioactivity"] = tmp
                data_out["SpecificRadioactivityUnits"] = "Bq/g"

    if InjectedRadioactivity and SpecificRadioactivity:
        data_out["InjectedRadioactivity"] = InjectedRadioactivity
        data_out["InjectedRadioactivityUnits"] = "MBq"
        data_out["SpecificRadioactivity"] = SpecificRadioactivity
        data_out["SpecificRadioactivityUnits"] = "Bq/g"
        numeric_check = [
            helper_functions.is_numeric(str(InjectedRadioactivity)),
            helper_functions.is_numeric(str(SpecificRadioactivity)),
        ]
        if False in numeric_check:
            data_out["InjectedMass"] = "n/a"
            data_out["InjectedMassUnits"] = "n/a"
        else:
            tmp = (InjectedRadioactivity * (10**6) / SpecificRadioactivity) * (10**6)
            if InjectedMass:
                if InjectedMass != tmp:
                    logger.warning(
                        "Inferred InjectedMass in ug doesn't match InjectedRadioactivity and "
                        "InjectedMass, could be a unit issue"
                    )
                data_out["InjectedMass"] = InjectedMass
                data_out["InjectedMassUnits"] = kwargs.get("InjectedMassUnits", "n/a")
            else:
                data_out["InjectedMass"] = tmp
                data_out["InjectedMassUnits"] = "ug"

    if InjectedMass and SpecificRadioactivity:
        data_out["InjectedMass"] = InjectedMass
        data_out["InjectedMassUnits"] = "ug"
        data_out["SpecificRadioactivity"] = SpecificRadioactivity
        data_out["SpecificRadioactivityUnits"] = "Bq/g"
        numeric_check = [
            helper_functions.is_numeric(str(SpecificRadioactivity)),
            helper_functions.is_numeric(str(InjectedMass)),
        ]
        if False in numeric_check:
            data_out["InjectedRadioactivity"] = "n/a"
            data_out["InjectedRadioactivityUnits"] = "n/a"
        else:
            tmp = ((InjectedMass / (10**6)) * SpecificRadioactivity) / (
                10**6
            )  # ((ug / 10 ^ 6) / Bq / g)/10 ^ 6 = MBq
            if InjectedRadioactivity:
                if InjectedRadioactivity != tmp:
                    logger.warning(
                        "Inferred InjectedRadioactivity in MBq doesn't match SpecificRadioactivity "
                        "and InjectedMass, could be a unit issue"
                    )
                data_out["InjectedRadioactivity"] = InjectedRadioactivity
                data_out["InjectedRadioactivityUnits"] = kwargs.get(
                    "InjectedRadioactivityUnits", "n/a"
                )
            else:
                data_out["InjectedRadioactivity"] = tmp
                data_out["InjectedRadioactivityUnits"] = "MBq"

    if MolarActivity and MolecularWeight:
        data_out["MolarActivity"] = MolarActivity
        data_out["MolarActivityUnits"] = "GBq/umol"
        data_out["MolecularWeight"] = MolecularWeight
        data_out["MolecularWeightUnits"] = "g/mol"
        numeric_check = [
            helper_functions.is_numeric(str(MolarActivity)),
            helper_functions.is_numeric(str(MolecularWeight)),
        ]
        if False in numeric_check:
            data_out["SpecificRadioactivity"] = "n/a"
            data_out["SpecificRadioactivityUnits"] = "n/a"
        else:
            tmp = (
                MolarActivity * (10**3)
            ) / MolecularWeight  # (GBq / umol * 10 ^ 6) / (g / mol / * 10 ^ 6) = Bq / g
            if SpecificRadioactivity:
                if SpecificRadioactivity != tmp:
                    logger.warning(
                        "Inferred SpecificRadioactivity in MBq/ug doesn't match Molar Activity and Molecular "
                        "Weight, could be a unit issue"
                    )
                data_out["SpecificRadioactivity"] = SpecificRadioactivity
                data_out["SpecificRadioactivityUnits"] = kwargs.get(
                    "SpecificRadioactivityUnityUnits", "n/a"
                )
            else:
                data_out["SpecificRadioactivity"] = tmp
                data_out["SpecificRadioactivityUnits"] = "Bq/g"

    if MolarActivity and SpecificRadioactivity:
        data_out["SpecificRadioactivity"] = SpecificRadioactivity
        data_out["SpecificRadioactivityUnits"] = "MBq/ug"
        data_out["MolarActivity"] = MolarActivity
        data_out["MolarActivityUnits"] = "GBq/umol"
        numeric_check = [
            helper_functions.is_numeric(str(SpecificRadioactivity)),
            helper_functions.is_numeric(str(MolarActivity)),
        ]
        if False in numeric_check:
            data_out["MolecularWeight"] = "n/a"
            data_out["MolecularWeightUnits"] = "n/a"
        else:
            tmp = (
                MolarActivity * 1000
            ) / SpecificRadioactivity  # (MBq / ug / 1000) / (GBq / umol) = g / mol
            if MolecularWeight:
                if MolecularWeight != tmp:
                    logger.warning(
                        "Inferred MolecularWeight in MBq/ug doesn't match Molar Activity and "
                        "Molecular Weight, could be a unit issue"
                    )

                data_out["MolecularWeight"] = tmp
                data_out["MolecularWeightUnits"] = kwargs.get(
                    "MolecularWeightUnits", "n/a"
                )
            else:
                data_out["MolecularWeight"] = tmp
                data_out["MolecularWeightUnits"] = "g/mol"

    if MolecularWeight and SpecificRadioactivity:
        data_out["SpecificRadioactivity"] = SpecificRadioactivity
        data_out["SpecificRadioactivityUnits"] = "MBq/ug"
        data_out["MolecularWeight"] = MolarActivity
        data_out["MolecularWeightUnits"] = "g/mol"
        numeric_check = [
            helper_functions.is_numeric(str(SpecificRadioactivity)),
            helper_functions.is_numeric(str(MolecularWeight)),
        ]
        if False in numeric_check:
            data_out["MolarActivity"] = "n/a"
            data_out["MolarActivityUnits"] = "n/a"
        else:
            tmp = MolecularWeight * (
                SpecificRadioactivity / 1000
            )  # g / mol * (MBq / ug / 1000) = GBq / umol
            if MolarActivity:
                if MolarActivity != tmp:
                    logger.warning(
                        "Inferred MolarActivity in GBq/umol doesn't match Specific Radioactivity and "
                        "Molecular Weight, could be a unit issue"
                    )
                data_out["MolarActivity"] = MolarActivity
                data_out["MolarActivityUnits"] = kwargs.get("MolarActivityUnits", "n/a")
            else:
                data_out["MolarActivity"] = tmp
                data_out["MolarActivityUnits"] = "GBq/umol"

    return data_out


def get_metadata_from_spreadsheet(
    metadata_path: Union[str, Path],
    image_folder,
    image_header_dict={},
    **additional_arguments,
) -> dict:
    """
    Extracts metadata from a spreadsheet and returns a dictionary of metadata organized under
    three main keys: nifti_json, blood_json, and blood_tsv

    :param metadata_path: path to a spreadsheet
    :type metadata_path: [str, pathlib.Path]
    :param image_folder: path to image folder
    :type image_folder: [str, pathlib.Path]
    :param image_header_dict: dictionary of image header values
    :type image_header_dict: dict
    :param additional_arguments: additional arguments to pass on, typically user sourced key value pairs
    :type additional_arguments: dict
    :return: dictionary of metadata
    :rtype: dict
    """
    spreadsheet_metadata = {"nifti_json": {}, "blood_json": {}, "blood_tsv": {}}
    spreadsheet_values = {}
    if Path(metadata_path).is_file():
        spreadsheet_values = helper_functions.single_spreadsheet_reader(
            path_to_spreadsheet=metadata_path,
            dicom_metadata=image_header_dict,
            **additional_arguments,
        )

    if Path(metadata_path).is_dir() or metadata_path == "":
        # we accept folder input as well as no input, in the
        # event of no input we search for spreadsheets in the
        # image folder
        if metadata_path == "":
            metadata_path = image_folder

        spreadsheets = helper_functions.collect_spreadsheets(metadata_path)
        pet_spreadsheets = [
            spreadsheet for spreadsheet in spreadsheets if is_pet.pet_file(spreadsheet)
        ]
        spread_sheet_values = {}

        for pet_spreadsheet in pet_spreadsheets:
            spreadsheet_values.update(
                helper_functions.single_spreadsheet_reader(
                    path_to_spreadsheet=pet_spreadsheet,
                    dicom_metadata=image_header_dict,
                    **additional_arguments,
                )
            )

        # remove any dates from the spreadsheet time values
    for key, value in spreadsheet_values.items():
        if "time" in key.lower():
            if isinstance(value, str):
                # check to see if the value converts to a datetime object with a date
                try:
                    time_value = parser.parse(value).time().strftime("%H:%M:%S")
                    spreadsheet_values[key] = time_value
                except ValueError:
                    pass
            elif isinstance(value, datetime.datetime):
                spreadsheet_values[key] = value.time().strftime("%H:%M:%S")
            elif isinstance(value, Timestamp):
                value = value.to_pydatetime()
                spreadsheet_values[key] = value.time().strftime("%H:%M:%S")
            else:
                pass

    # check for any blood (tsv) data or otherwise in the given spreadsheet values
    blood_tsv_columns = [
        "time",
        "plasma_radioactivity",
        "metabolite_parent_fraction",
        "whole_blood_radioactivity",
    ]
    blood_json_columns = [
        "PlasmaAvail",
        "WholeBloodAvail",
        "MetaboliteAvail",
        "MetaboliteMethod",
        "MetaboliteRecoveryCorrectionApplied",
        "DispersionCorrected",
    ]

    # check for existing tsv columns
    for column in blood_tsv_columns:
        try:
            values = spreadsheet_values[column]
            spreadsheet_metadata["blood_tsv"][column] = values
            # pop found data from spreadsheet values after it's been found
            spreadsheet_values.pop(column)
        except KeyError:
            pass

    # even out the values in the blood tsv columns if they're different lengths by appending zeros to the end
    # of each column/list
    # determine the longest column
    column_lengths = [
        len(column) for column in spreadsheet_metadata["blood_tsv"].values()
    ]

    try:
        longest_column = max(column_lengths)
    except ValueError:
        # columns are all the same length or there are no columns
        longest_column = None
    if longest_column:
        # iterate over each column, determine how many zeros to append to the end of each column
        for column in spreadsheet_metadata["blood_tsv"].keys():
            zeros_to_append = longest_column - len(
                spreadsheet_metadata["blood_tsv"][column]
            )
            spreadsheet_metadata["blood_tsv"][column] += [0] * zeros_to_append

    # check for existing blood json values
    for column in blood_json_columns:
        try:
            values = spreadsheet_values[column]
            spreadsheet_metadata["blood_json"][column] = values
            # pop found data from spreadsheet values after it's been found
            spreadsheet_values.pop(column)
        except KeyError:
            pass

    if not spreadsheet_metadata.get("nifti_json", None):
        spreadsheet_metadata["nifti_json"] = {}
    spreadsheet_metadata["nifti_json"].update(spreadsheet_values)

    return spreadsheet_metadata


if __name__ == "__main__":
    update_json_cli()



"""
Contains class ecat2ii, used to write ecat information out to a nifti file, depends on
`nibabel <https://nipy.org/nibabel/>`_ and  :func:`pypet2bids.read_ecat.read_ecat`.

| *Authors: Anthony Galassi*
| *Copyright OpenNeuroPET team*
"""

import datetime
import nibabel
import numpy
import pathlib
from pypet2bids.read_ecat import (
    read_ecat,
    code_dir,
)  # we use code_dir for ecat debugging
import os
import pickle
import logging

try:
    import helper_functions
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions


# debug variable
# save steps for debugging, for more info see ecat_testing/README.md
ecat_save_steps = os.environ.get("ECAT_SAVE_STEPS", 0)
if ecat_save_steps == "1":
    # check to see if the code directory is available, if it's not create it and
    # the steps dir to save outputs created if ecat_save_steps is set to 1
    steps_dir = code_dir.parent / "ecat_testing" / "steps"
    if not steps_dir.is_dir():
        os.makedirs(code_dir.parent / "ecat_testing" / "steps", exist_ok=True)

logger = logging.getLogger("pypet2bids")


def ecat2nii(
    ecat_main_header=None,
    ecat_subheaders=None,
    ecat_pixel_data=None,
    ecat_file=None,
    nifti_file: str = "",
    sif_out=False,
    affine=None,
    save_binary=False,
    **kwargs,
):
    """
    Converts an ECAT file into a nifti and a sidecar json, used in conjunction with
    :func:`pypet2bids.read_ecat.read_ecat`

    :param ecat_main_header: the main header of an ECAT file
    :param ecat_subheaders: the subheaders for each frame of the ECAT file
    :param ecat_pixel_data: the imaging/pixel data from the ECAT file
    :param ecat_file: the path to the ECAT file, required and used to create .nii and .json output files
    :param nifti_file: the desired output path of the nifti file
    :param sif_out: outputs a .sif file containing the images pixel data
    :param affine: a user supplied affine, this is gathered from the ECAT if not supplied.
    :param save_binary: dumps a pickled ECAT object, you probably shouldn't be using this.
    :param kwargs: additional key value pairs that one wishes to add to the sidecar json accompanying a converted ECAT
        nii image
    :return: a nibabel nifti object if one wishes to muddle with the object in python and not in a .nii file

    """

    # if a nifti file/path is not included write a nifti next to the ecat file
    if not nifti_file:
        nifti_file = os.path.splitext(ecat_file)[0] + ".nii"
    else:
        nifti_file = nifti_file

    if not pathlib.Path(nifti_file).parent.exists():
        pathlib.Path(nifti_file).parent.mkdir(parents=True, exist_ok=True)

    # collect the output folder from the nifti path will use for .sif files
    output_folder = pathlib.Path(nifti_file).parent
    nifti_file_w_out_extension = os.path.splitext(str(pathlib.Path(nifti_file).name))[0]

    # if already read nifti file skip re-reading
    if (
        ecat_main_header is None
        and ecat_subheaders is None
        and ecat_pixel_data is None
        and ecat_file
    ):
        # collect ecat_file
        main_header, sub_headers, data = read_ecat(ecat_file=ecat_file)
    elif (
        ecat_file is None
        and type(ecat_main_header) is dict
        and type(ecat_subheaders) is list
        and type(ecat_pixel_data) is numpy.ndarray
    ):
        main_header, sub_headers, data = (
            ecat_main_header,
            ecat_subheaders,
            ecat_pixel_data,
        )
    else:
        raise Exception(
            "Must pass in filepath for ECAT file or "
            "(ecat_main_header, ecat_subheaders, and ecat_pixel data "
            f"got ecat_file={ecat_file}, type(ecat_main_header)={type(ecat_main_header)}, "
            f"type(ecat_subheaders)={type(ecat_subheaders)}, "
            f"type(ecat_pixel_data)={type(ecat_pixel_data)} instead."
        )

    # debug step #6 view data as passed to ecat2nii method
    if ecat_save_steps == "1":
        # collect only the first, middle, and last frames from pixel_data_matrix_4d as first, middle, and last
        # frames are typically the most interesting
        frames = [0, len(data) // 2, -1]
        frames_to_record = []
        for f in frames:
            frames_to_record.append(data[:, :, :, f])

        # now collect a single 2d slice from the "middle" of the 3d frames in frames_to_record
        slice_to_record = []
        for index, frame in enumerate(frames_to_record):
            numpy.savetxt(
                steps_dir / f"6_ecat2nii_python_{index}.tsv",
                frames_to_record[index][:, :, frames_to_record[index].shape[2] // 2],
                delimiter="\t",
                fmt="%s",
            )

        helper_functions.first_middle_last_frames_to_text(
            four_d_array_like_object=data,
            output_folder=steps_dir,
            step_name="6_ecat2nii_python",
        )

    # set the byte order and the pixel data type from the input array
    pixel_data_type = data.dtype

    # check for TimeZero supplied via kwargs
    if kwargs.get("TimeZero", None):
        TimeZero = kwargs["TimeZero"]
    else:
        logger.warning(
            "Metadata TimeZero is missing -- set to ScanStart or empty to use the scanning time as "
            "injection time"
        )

    # get image shape
    img_shape = data.shape
    shape_from_headers = (
        sub_headers[0]["X_DIMENSION"],
        sub_headers[0]["Y_DIMENSION"],
        sub_headers[0]["Z_DIMENSION"],
        main_header["NUM_FRAMES"],
    )

    # make sure number of data elements matches frame number
    single_frame = False
    if img_shape[3] == 1 and img_shape[0:2] == shape_from_headers[0:2]:
        single_frame = True
    if img_shape != shape_from_headers and not single_frame:
        raise Exception(
            f"Mismatch between expected X,Y,Z, and Num. Frames dimensions ({shape_from_headers} obtained from headers"
            f"and shape of imaging data ({img_shape}"
        )

    # format data into acceptable shape for nibabel, by first creating empty matrix
    img_temp = numpy.zeros(
        shape=(
            sub_headers[0]["X_DIMENSION"],
            sub_headers[0]["Y_DIMENSION"],
            sub_headers[0]["Z_DIMENSION"],
            main_header["NUM_FRAMES"],
        ),
        dtype=">f4",
    )

    # collect timing information
    start, delta = [], []

    # collect prompts and randoms
    prompts, randoms = [], []

    # load frame data into img temp
    for index in reversed(
        range(img_shape[3])
    ):  # Don't throw stones working from existing matlab code
        print(f"Loading frame {index + 1}")
        img_temp[:, :, :, index] = numpy.flip(
            numpy.flip(
                numpy.flip(
                    data[:, :, :, index] * sub_headers[index]["SCALE_FACTOR"], 1
                ),
                2,
            ),
            0,
        )
        start.append(sub_headers[index]["FRAME_START_TIME"] * 60)  # scale to per minute
        delta.append(sub_headers[index]["FRAME_DURATION"] * 60)  # scale to per minute

        if main_header.get("SW_VERSION", 0) >= 73:
            # scale both to per minute
            prompts.append(
                sub_headers[index]["PROMPT_RATE"]
                * sub_headers[index]["FRAME_DURATION"]
                * 60
            )
            randoms.append(
                sub_headers[index]["RANDOM_RATE"]
                * sub_headers[index]["FRAME_DURATION"]
                * 60
            )
        else:
            # this field is not available in ecat 7.2
            prompts.append(0)
            randoms.append(0)

    # debug step #7 view data after flipping into nifti space/orientation
    if ecat_save_steps == "1":
        helper_functions.first_middle_last_frames_to_text(
            four_d_array_like_object=img_temp,
            output_folder=steps_dir,
            step_name="7_flip_ecat2nii_python",
        )

    # so the only real difference between the matlab code and the python code is that that we aren't manually
    # scaling the date to 16 bit integers.
    rg = img_temp.max() - img_temp.min()
    if rg != 32767:
        max_img = img_temp.max()
        img_temp = img_temp / max_img * 32767
        sca = max_img / 32767
        min_img = img_temp.min()
        if min_img < -32768:
            img_temp = img_temp / (min_img * -32768)
            sca = sca * (min_img * -32768)
    if ecat_save_steps == "1":
        with open(os.path.join(steps_dir, "8.5_sca.txt"), "w") as sca_file:
            sca_file.write(f"Scaling factor: {sca}\n")
            sca_file.write(
                f"Scaling factor * ECAT Cal Factor: {sca * main_header['ECAT_CALIBRATION_FACTOR']}\n"
            )

    # scale image to 16 bit
    final_image = img_temp.astype(numpy.single)

    # debug step 8 check after "rescaling" to 16 bit
    if ecat_save_steps == "1":
        helper_functions.first_middle_last_frames_to_text(
            four_d_array_like_object=final_image,
            output_folder=steps_dir,
            step_name="8_rescale_to_16_ecat2nii_python",
        )

    ecat_cal_units = main_header[
        "CALIBRATION_UNITS"
    ]  # Header field designating whether data has already been calibrated
    if ecat_cal_units == 1:  # Calibrate if it hasn't been already
        final_image = (
            numpy.round(final_image) * main_header["ECAT_CALIBRATION_FACTOR"] * sca
        )
        # this debug step may not execute if we're not calibrating the scan, but that's okay
        if ecat_save_steps == "1":
            helper_functions.first_middle_last_frames_to_text(
                four_d_array_like_object=final_image,
                output_folder=steps_dir,
                step_name="9_scal_cal_units_ecat2nii_python",
            )
    else:  # And don't calibrate if CALIBRATION_UNITS is anything else but 1
        final_image = numpy.round(final_image) * sca

    qoffset_x = -1 * (
        (
            (sub_headers[0]["X_DIMENSION"] * sub_headers[0]["X_PIXEL_SIZE"] * 10 / 2)
            - sub_headers[0]["X_PIXEL_SIZE"] * 5
        )
    )

    qoffset_y = -1 * (
        (
            (sub_headers[0]["Y_DIMENSION"] * sub_headers[0]["Y_PIXEL_SIZE"] * 10 / 2)
            - sub_headers[0]["Y_PIXEL_SIZE"] * 5
        )
    )

    qoffset_z = -1 * (
        (
            (sub_headers[0]["Z_DIMENSION"] * sub_headers[0]["Z_PIXEL_SIZE"] * 10 / 2)
            - sub_headers[0]["Z_PIXEL_SIZE"] * 5
        )
    )

    # build affine if it's not included in function call
    if not affine:
        mat = (
            numpy.diag(
                [
                    sub_headers[0]["X_PIXEL_SIZE"],
                    sub_headers[0]["Y_PIXEL_SIZE"],
                    sub_headers[0]["Z_PIXEL_SIZE"],
                ]
            )
            * 10
        )
        affine = nibabel.affines.from_matvec(mat, [qoffset_x, qoffset_y, qoffset_z])

    img_nii = nibabel.Nifti1Image(final_image, affine=affine)

    # debug step 10, check to see what's happened after we've converted our numpy array in to a nibabel object
    if ecat_save_steps == "1":
        helper_functions.first_middle_last_frames_to_text(
            four_d_array_like_object=img_nii.dataobj,
            output_folder=steps_dir,
            step_name="10_save_nii_ecat2nii_python",
        )

    img_nii.header.set_slope_inter(slope=1, inter=0)
    img_nii.header.set_xyzt_units("mm", "sec")
    img_nii.header.set_qform(affine, code=1)
    img_nii.header.set_sform(affine, code=1)
    # No setter methods for these
    img_nii.header["cal_max"] = final_image.max()
    img_nii.header["cal_min"] = final_image.min()
    img_nii.header["descrip"] = "OpenNeuroPET ecat2nii.py conversion"

    nibabel.save(img_nii, nifti_file)

    # run step 11 in debug
    if ecat_save_steps == "1":
        # load nifti file with nibabel
        written_img_nii = nibabel.load(nifti_file)

        helper_functions.first_middle_last_frames_to_text(
            four_d_array_like_object=written_img_nii.dataobj,
            output_folder=steps_dir,
            step_name="11_read_saved_nii_python",
        )

    # used for testing veracity of nibabel read and write.
    if save_binary:
        pickle.dump(img_nii, open(nifti_file + ".pickle", "wb"))

    # write out timing file
    if sif_out:
        with open(
            os.path.join(output_folder, nifti_file_w_out_extension + ".sif"), "w"
        ) as sif_file:
            scantime = datetime.datetime.fromtimestamp(main_header["SCAN_START_TIME"])
            scantime = scantime.astimezone().isoformat()
            sif_file.write(f"{scantime} {len(start)} 4 1\n")
            for index in reversed(range(len(start))):
                start_i = round(start[index])
                start_i_plus_delta_i = start_i + round(delta[index])
                prompt = round(prompts[index])
                random = round(randoms[index])
                output_string = f"{start_i} {start_i_plus_delta_i} {prompt} {random}\n"
                sif_file.write(output_string)

    return img_nii



import numpy
import dotenv
from pypet2bids.write_ecat import write_ecat
from pypet2bids.read_ecat import read_ecat, ecat_header_maps
import os
from math import e
from pathlib import Path
from scipy.io import savemat

"""
This script exists solely to create an ecat with pixel values ranging from 0 to 32767 with a minimum number of frames.
This evenly pixel spaced ecat that has a small number of frames will be used to evaluate the accuracy and precision
of a number of tools that convert or otherwise manipulate ecat data by supplying a ecat image and a corresponding text
file containing the pixel data of that ecat.

This ECAT file and script is thus referred to as the Golden ECAT b/c it represents the perfect standard by which to 
judge all ECATs.

The accompanying text file is formatted such that the pixel values are written one per line:

0
1
2
3
.
.
.
32767

These values are then transformed into a set of NxNxNxF where N = the dimension of a matrix, and F = the total number of
frames within the ecat. 

Anthony Galassi - 2022
----------------------------------------------
Copyright Open NeuroPET team
"""


def main():
    script_path = Path(__file__).absolute()
    ecat_validation_folder = os.path.join(
        script_path.parent.parent.parent, "ecat_validation/"
    )

    # Path to a reference/skeleton ecat as well as output paths for saving the created ecats to are stored in
    # environment variables or a .env file
    dotenv.load_dotenv(dotenv.find_dotenv())

    # collect path to golden ecat file
    int_golden_ecat_path = os.environ["GOLDEN_ECAT_INTEGER"]
    int_golden_ecat_path_stem = Path(int_golden_ecat_path).stem

    # collect skeleton header and subheaders.
    skeleton_main_header, skeleton_subheader, _ = read_ecat(
        os.environ["GOLDEN_ECAT_TEMPLATE_ECAT"], collect_pixel_data=False
    )
    number_of_frames = 4
    one_dimension = 16
    # generate known 'pixel' data e.g. 4 frames with volume = 4x4x4
    number_of_array_elements = (one_dimension**3) * number_of_frames

    # first generate integer pixel data
    # 0 to 32767
    image_min = 0
    image_max = 32767
    spacing = round((image_max - image_min) / number_of_array_elements)

    # integer_pixel_data = numpy.arange(image_min, image_max, dtype=numpy.ushort, step=spacing)
    integer_pixel_data = numpy.arange(image_min, image_max, dtype=">H", step=spacing)

    # save data for analysis in matlab
    matlab_struct = {}

    # reshape pixel data into 3-d arrays
    pixels_to_collect = one_dimension**3
    frames = []
    temp_three_d_arrays = numpy.array_split(integer_pixel_data, number_of_frames)
    for i in range(number_of_frames):
        frames.append(
            temp_three_d_arrays[i].reshape(one_dimension, one_dimension, one_dimension)
        )
        matlab_struct[f"frame_{i + 1}_pixel_data"] = frames[i]

    # edit the header to suit the new file
    header_to_write = skeleton_main_header
    header_to_write["NUM_FRAMES"] = 4
    header_to_write["ORIGINAL_FILE_NAME"] = "GoldenECATInteger"
    header_to_write["STUDY_TYPE"] = "Golden"
    header_to_write["PATIENT_ID"] = "PerfectPatient"
    header_to_write["PATIENT_NAME"] = "Majesty"
    header_to_write["FACILITY_NAME"] = "Virtual"
    header_to_write["NUM_PLANES"] = one_dimension
    header_to_write["ECAT_CALIBRATION_FACTOR"] = 1.0

    subheaders_to_write = skeleton_subheader[0:number_of_frames]
    for subheader in subheaders_to_write:
        subheader["X_DIMENSION"] = (
            one_dimension  # pixel data is 3-d this is 1/3 root of the 3d array.
        )
        subheader["Y_DIMENSION"] = one_dimension
        subheader["Z_DIMENSION"] = one_dimension
        subheader["IMAGE_MIN"] = integer_pixel_data.min()
        subheader["IMAGE_MAX"] = integer_pixel_data.max()
        subheader["ANNOTATION"] = "This patient is very small."
        subheader["DATA_TYPE"] = 6
        subheader["SCALE_FACTOR"] = 1

    matlab_struct["subheaders"] = subheaders_to_write
    matlab_struct["mainheader"] = header_to_write

    write_ecat(
        ecat_file=int_golden_ecat_path,
        mainheader_schema=ecat_header_maps["ecat_headers"]["73"]["mainheader"],
        mainheader_values=header_to_write,
        subheaders_values=subheaders_to_write,
        subheader_schema=ecat_header_maps["ecat_headers"]["73"]["7"],
        number_of_frames=number_of_frames,
        pixel_x_dimension=one_dimension,
        pixel_y_dimension=one_dimension,
        pixel_z_dimension=one_dimension,
        pixel_byte_size=2,
        pixel_data=frames,
    )

    savemat(
        os.path.join(ecat_validation_folder, (int_golden_ecat_path_stem + ".mat")),
        matlab_struct,
    )

    # now read it just to make sure what was created is a real file and not error riddled.
    (
        int_golden_ecat_main_header,
        int_golden_ecat_subheaders,
        int_golden_ecat_pixel_data,
    ) = read_ecat(int_golden_ecat_path)


if __name__ == "__main__":
    main()



"""
This module contains the Ecat class which uses collections of other functions in this library to read native ECAT files
and write them out to Nifti files.


| *Anthony Galassi*
| *Copyright Open NeuroPET team*
"""

import datetime
import re
import nibabel
import os
import json
import pathlib
import pandas as pd

try:
    import helper_functions
    import sidecar
    import read_ecat
    import ecat2nii
    import dcm2niix4pet
    from update_json_pet_file import (
        get_metadata_from_spreadsheet,
        check_meta_radio_inputs,
    )
    from telemetry import telemetry_enabled, send_telemetry
except ModuleNotFoundError:
    import pypet2bids.helper_functions as helper_functions
    import pypet2bids.sidecar as sidecar
    import pypet2bids.read_ecat as read_ecat
    import pypet2bids.ecat2nii as ecat2nii
    import pypet2bids.dcm2niix4pet as dcm2niix4pet
    from pypet2bids.update_json_pet_file import (
        get_metadata_from_spreadsheet,
        check_meta_radio_inputs,
    )
    from pypet2bids.telemetry import telemetry_enabled, send_telemetry

from dateutil import parser

logger = helper_functions.logger("pypet2bids")


def parse_this_date(date_like_object) -> str:
    """
    Uses the `dateutil.parser` module to extract a date from a variety of differently formatted date strings

    :param date_like_object: something that resembles a timestamp or a date time, could be integer, float, or string.
    :return: an hour minute second datetime string.
    """
    if type(date_like_object) is int:
        parsed_date = datetime.datetime.fromtimestamp(date_like_object)
    else:
        parsed_date = parser.parse(date_like_object)

    return parsed_date.strftime("%H:%M:%S")


class Ecat:
    """
    This class reads an ecat file -> extracts header, subheader, and image matrices for
    viewing in stdout. Additionally, this class can be used to convert an ECAT7.X image into a nifti image.
    """

    def __init__(
        self,
        ecat_file,
        nifti_file=None,
        decompress=True,
        collect_pixel_data=True,
        metadata_path=None,
        kwargs={},
        ezbids=False,
    ):
        """
        Initialization of this class requires only a path to an ecat file.

        :param ecat_file: path to a valid ecat file
        :param nifti_file: when using this class for conversion from ecat to nifti this path, if supplied, will be used
            to output the newly generated nifti
        :param decompress: attempt to decompress the ecat file, should probably be set to false
        """
        self.ecat_header = {}  # ecat header information is stored here
        self.subheaders = []  # subheader information is placed here
        self.ecat_info = {}
        self.affine = {}  # affine matrix/information is stored here.
        self.frame_start_times = (
            []
        )  # frame_start_times, frame_durations, and decay_factors are all
        self.frame_durations = (
            []
        )  # extracted from ecat subheaders. They're pretty important and get
        self.decay_factors = []  # stored here
        self.sidecar_template = (
            sidecar.sidecar_template_full
        )  # bids approved sidecar file with ALL bids fields
        self.sidecar_template_short = (
            sidecar.sidecar_template_short
        )  # bids approved sidecar with only required bids fields
        self.sidecar_path = None
        self.directory_table = None
        self.spreadsheet_metadata = {
            "nifti_json": {},
            "blood_tsv": {},
            "blood_json": {},
        }
        self.kwargs = kwargs
        self.output_path = None
        self.metadata_path = metadata_path
        self.ezbids = ezbids

        self.telemetry_data = {}

        # load config file
        default_json_path = helper_functions.check_pet2bids_config(
            "DEFAULT_METADATA_JSON"
        )
        if default_json_path and pathlib.Path(default_json_path).exists():
            with open(default_json_path, "r") as json_file:
                try:
                    self.spreadsheet_metadata.update(json.load(json_file))
                except json.decoder.JSONDecodeError:
                    logger.warning(
                        f"Unable to load default metadata json file at {default_json_path}, skipping."
                    )

        if os.path.isfile(ecat_file):
            self.ecat_file = str(ecat_file)
        else:
            raise FileNotFoundError(ecat_file)

        if ".gz" in self.ecat_file and decompress is True:
            uncompressed_ecat_file = re.sub(".gz", "", self.ecat_file)
            helper_functions.decompress(self.ecat_file, uncompressed_ecat_file)
            self.ecat_file = uncompressed_ecat_file

        if ".gz" in self.ecat_file and decompress is False:
            raise Exception("Nifti must be decompressed for reading of file headers")

        try:
            self.ecat = nibabel.ecat.load(self.ecat_file)
        except nibabel.filebasedimages.ImageFileError as err:
            print("\nFailed to load ecat image.\n")
            raise err

        directory_byte_block = read_ecat.read_bytes(
            path_to_bytes=self.ecat_file, byte_start=512, byte_stop=1024
        )

        self.directory_table = read_ecat.get_directory_data(
            directory_byte_block, self.ecat_file
        )

        # extract ecat info
        self.extract_affine()
        if collect_pixel_data:
            self.ecat_header, self.subheaders, self.data = read_ecat.read_ecat(
                self.ecat_file
            )
        else:
            self.ecat_header, self.subheaders, self.data = read_ecat.read_ecat(
                self.ecat_file, collect_pixel_data=False
            )

        # aggregate ecat info into ecat_info dictionary
        self.ecat_info["header"] = self.ecat_header
        self.ecat_info["subheaders"] = self.subheaders
        self.ecat_info["affine"] = self.affine

        # swap file extensions and save output nifti with same name as original ecat
        if not nifti_file:
            self.nifti_file = os.path.splitext(self.ecat_file)[0] + ".nii"
        else:
            self.nifti_file = nifti_file

        self.telemetry_data["InputType"] = "ECAT" + str(self.ecat_header["SW_VERSION"])
        self.telemetry_data["TotalInputFiles"] = 1
        self.telemetry_data["TotalInputFilesSize"] = (
            pathlib.Path(self.ecat_file).stat().st_size
        )

        # que up metadata path for spreadsheet loading later
        if self.metadata_path:
            if (
                pathlib.Path(metadata_path).is_file()
                and pathlib.Path(metadata_path).exists()
            ):
                self.metadata_path = metadata_path
        elif metadata_path == "":
            self.metadata_path = pathlib.Path(self.ecat_file).parent
        else:
            self.metadata_path = None

        if self.metadata_path:
            load_spreadsheet_data = get_metadata_from_spreadsheet(
                metadata_path=self.metadata_path,
                image_folder=pathlib.Path(self.ecat_file).parent,
                image_header_dict={},
            )

            self.spreadsheet_metadata["nifti_json"].update(
                load_spreadsheet_data["nifti_json"]
            )
            self.spreadsheet_metadata["blood_tsv"].update(
                load_spreadsheet_data["blood_tsv"]
            )
            self.spreadsheet_metadata["blood_json"].update(
                load_spreadsheet_data["blood_json"]
            )

            if helper_functions.collect_spreadsheets(self.metadata_path):
                self.telemetry_data.update({"metadata_spreadsheet_user": True})
            else:
                self.telemetry_data.update({"metadata_spreadsheet_user": False})

            if self.spreadsheet_metadata.get("blood_tsv", None):
                self.telemetry_data.update({"blood_tsv": True})
            else:
                self.telemetry_data.update({"blood_tsv": False})

    def make_nifti(self, output_path=None):
        """
        Outputs a nifti from the read in ECAT file.
        :param output_path: Optional path to output file to, if not supplied saves nifti in same directory as ECAT
        :param output_path: Optional path to output file to, if not supplied saves nifti in same directory as ECAT
        :type output_path:
        :return: the output path the nifti was written to, used later for placing metadata/sidecar files
        :rtype:
        """

        # save nifti
        if output_path is None:
            output = self.nifti_file
        else:
            output = output_path
        ecat2nii.ecat2nii(
            ecat_main_header=self.ecat_header,
            ecat_subheaders=self.subheaders,
            ecat_pixel_data=self.data,
            nifti_file=output,
            affine=self.affine,
        )

        self.telemetry_data["NiftiFiles"] = 1
        self.telemetry_data["NiftiFilesSize"] = pathlib.Path(output).stat().st_size

        if "nii.gz" not in pathlib.Path(output).name:
            output = helper_functions.compress(output)

        return output

    def extract_affine(self):
        """
        Extract affine matrix from ecat
        """
        self.affine = self.ecat.affine.tolist()

    def show_affine(self):
        """
        Display affine to stdout
        :return: affine matrix row by row.
        """
        for row in self.affine:
            print(row)

    def show_directory_table(self):
        """
        Prints the directory table for the ECAT file to stdout.

        :return: None
        """
        for row in range(self.directory_table.shape[0]):
            for column in range(self.directory_table.shape[1]):
                if column == self.directory_table.shape[1] - 1:
                    print(self.directory_table[row][column])
                else:
                    print(self.directory_table[row][column], end=",", sep="")

    def show_header(self):
        """
        Display header to stdout in key: value format.

        :return: None
        """
        for key, value in self.ecat_header.items():
            print(f"{key}: {value}")

    def show_subheaders(self):
        """
        Displays subheaders to stdout.

        :return: None
        """
        for subheader in self.subheaders:
            print(subheader)

    def populate_sidecar(self, **kwargs):
        """
        Creates a side-car dictionary with any bids relevant information extracted from the ecat.

        :param kwargs: Populates sidecar file with relevant PET information, additional information that is not in the
            ECAT file can be supplied as a dictionary argument via kwargs.
        :return: None
        """

        # if it's an ecat it's Siemens
        self.sidecar_template["Manufacturer"] = "Siemens"
        # Siemens model best guess
        self.sidecar_template["ManufacturersModelName"] = self.ecat_header.get(
            "SERIAL_NUMBER", None
        )
        self.sidecar_template["TracerRadionuclide"] = self.ecat_header.get(
            "ISOTOPE_NAME", None
        )
        self.sidecar_template["PharmaceuticalName"] = self.ecat_header.get(
            "RADIOPHARMACEUTICAL", None
        )

        # collect frame time start and populate various subheader fields
        for subheader in self.subheaders:
            self.sidecar_template["DecayCorrectionFactor"].append(
                subheader.get("DECAY_CORR_FCTR", None)
            )
            self.sidecar_template["FrameTimesStart"].append(
                subheader.get("FRAME_START_TIME", None)
            )
            self.sidecar_template["FrameDuration"].append(
                subheader.get("FRAME_DURATION", None)
            )
            self.sidecar_template["ScaleFactor"].append(
                subheader.get("SCALE_FACTOR", None)
            )

            # note some of these values won't be in the subheaders for the standard matrix image
            # need to make sure to clean up arrays and fields filled w/ none during pruning
            if subheader.get("SCATTER_FRACTION", None):
                self.sidecar_template["ScatterFraction"].append(
                    subheader.get("SCATTER_FRACTION")
                )
            if subheader.get("PROMPT_RATE", None):
                self.sidecar_template["PromptRate"].append(subheader.get("PROMPT_RATE"))
            if subheader.get("RANDOM_RATE", None):
                self.sidecar_template["RandomRate"].append(subheader.get("RANDOM_RATE"))
            if subheader.get("SINGLES_RATE", None):
                self.sidecar_template["SinglesRate"].append(
                    subheader.get("SINGLES_RATE")
                )

        # collect possible reconstruction method from subheader
        recon_method = helper_functions.get_recon_method(
            self.subheaders[0].get("ANNOTATION")
        )
        if recon_method:
            self.sidecar_template.update(**recon_method)

        # collect and convert start times for acquisition/time zero?
        scan_start_time = self.ecat_header.get("SCAN_START_TIME", None)

        if scan_start_time:
            scan_start_time = parse_this_date(scan_start_time)
            self.sidecar_template["AcquisitionTime"] = scan_start_time
            self.sidecar_template["ScanStart"] = scan_start_time

        # collect dose start time
        dose_start_time = self.ecat_header.get("DOSE_START_TIME", None)
        if dose_start_time is not None:
            if (
                dose_start_time != 0
                and dose_start_time != "0"
                and dose_start_time != 0.0
            ):
                dose_start_time = parse_this_date(dose_start_time)
                parsed_dose_time = parse_this_date(dose_start_time)
                self.sidecar_template["PharmaceuticalDoseTime"] = parsed_dose_time
                self.sidecar_template["InjectionStart"] = parsed_dose_time
            else:
                self.sidecar_template["PharmaceuticalDoseTime"] = int(dose_start_time)
                self.sidecar_template["InjectionStart"] = int(dose_start_time)

        # if decay correction exists mark decay correction boolean as true
        if len(self.decay_factors) > 0:
            self.sidecar_template["ImageDecayCorrected"] = "true"

        # calculate scaling factor
        sca = self.data.max() / 32767

        self.sidecar_template["DoseCalibrationFactor"] = sca * self.ecat_header.get(
            "ECAT_CALIBRATION_FACTOR"
        )
        self.sidecar_template["Filename"] = os.path.basename(self.nifti_file)
        self.sidecar_template["ImageSize"] = [
            self.subheaders[0]["X_DIMENSION"],
            self.subheaders[0]["Y_DIMENSION"],
            self.subheaders[0]["Z_DIMENSION"],
            self.ecat_header["NUM_FRAMES"],
        ]

        self.sidecar_template["PixelDimensions"] = [
            self.subheaders[0]["X_PIXEL_SIZE"] * 10,
            self.subheaders[0]["Y_PIXEL_SIZE"] * 10,
            self.subheaders[0]["Z_PIXEL_SIZE"] * 10,
        ]

        # add tag for conversion software
        self.sidecar_template["ConversionSoftware"] = "pypet2bids"
        self.sidecar_template["ConversionSoftwareVersion"] = (
            helper_functions.get_version()
        )

        # update sidecar values from spreadsheet
        if self.spreadsheet_metadata.get("nifti_json", None):
            self.sidecar_template.update(self.spreadsheet_metadata["nifti_json"])

        # include any additional values
        if kwargs:
            self.sidecar_template.update(**kwargs)

        if not self.sidecar_template.get("TimeZero", None) and not kwargs.get(
            "TimeZero", None
        ):
            if not self.sidecar_template.get(
                "AcquisitionTime", None
            ) and not kwargs.get("TimeZero", None):
                logger.warn(
                    f"Unable to determine TimeZero for {self.ecat_file}, you need will need to provide this"
                    f" for a valid BIDS sidecar."
                )
            else:
                self.sidecar_template["TimeZero"] = self.sidecar_template[
                    "AcquisitionTime"
                ]

        # set scan start and pharmaceutical dose time relative to time zero.
        times_make_relative = ["ScanStart", "PharmaceuticalDoseTime", "InjectionStart"]
        time_zero_datetime = datetime.datetime.strptime(
            self.sidecar_template.get("TimeZero"), "%H:%M:%S"
        )
        for t in times_make_relative:
            t_value = self.sidecar_template.get(t)
            # sometimes we start with 0 or time in seconds, we first check for this
            try:
                int(t_value)
                self.sidecar_template[t] = t_value
            except ValueError:
                t_datetime = datetime.datetime.strptime(t_value, "%H:%M:%S")
                time_diff = t_datetime - time_zero_datetime
                self.sidecar_template[t] = time_diff.total_seconds()

        # clear any nulls from json sidecar and replace with none's
        self.sidecar_template = helper_functions.replace_nones(self.sidecar_template)

        # lastly infer radio data if we have it
        meta_radio_inputs = check_meta_radio_inputs(self.sidecar_template)
        self.sidecar_template.update(**meta_radio_inputs)

        # set ModeOfAdministration to lower case
        if self.sidecar_template.get("ModeOfAdministration", ""):
            self.sidecar_template["ModeOfAdministration"] = self.sidecar_template[
                "ModeOfAdministration"
            ].lower()

    def prune_sidecar(self):
        """
        Eliminate unpopulated fields in sidecar while leaving in mandatory fields even if they are unpopulated.

        :return: a list of removed fields from the sidecar file
        """
        short_fields = list(self.sidecar_template_short.keys())
        full_fields = list(self.sidecar_template)
        exclude_list = []
        for field, value in self.sidecar_template.items():
            if value is not None and value != "":
                # check to make sure value isn't a list of null types
                # e.g. if value = [None, None, None] we don't want to include it.
                if type(value) is list:
                    none_count = value.count(None)
                    if len(value) == none_count:
                        pass
                    else:
                        exclude_list.append(field)
                else:
                    exclude_list.append(field)

        exclude_list = exclude_list + short_fields

        destroy_list = set(full_fields) - set(exclude_list)

        destroyed = []
        for to_be_destroyed in destroy_list:
            destroyed.append(self.sidecar_template.pop(to_be_destroyed))

        return destroyed

    def show_sidecar(self, output_path=None):
        """
        Write sidecar file to a json or display to stdout if no filepath is supplied
        :param output_path: path to output a json file
        :return: None
        """
        self.prune_sidecar()
        self.sidecar_template = helper_functions.replace_nones(self.sidecar_template)

        # this is mostly for ezBIDS, but it helps us to make better use of the series description that
        # dcm2niix generates by default for PET imaging, here we mirror the dcm2niix output for ecats
        collect_these_fields = {
            "ProtocolName": "",
            "SeriesDescription": "",
            "TracerName": "trc",
            "InjectedRadioactivity": "",
            "InjectedRadioactivityUnits": "",
            "ReconMethodName": "rec",
            "TimeZero": "",
        }
        collection_of_fields = {}
        for field, entity_string in collect_these_fields.items():
            if self.sidecar_template.get(field, None):
                # if there's a shortened entity string for the field use that
                if entity_string != "":
                    collection_of_fields[entity_string] = self.sidecar_template.get(
                        field
                    )
                else:
                    collection_of_fields[field] = self.sidecar_template.get(field)
        if self.output_path:
            if helper_functions.collect_bids_part("ses", self.output_path) != "":
                collection_of_fields["ses"] = helper_functions.collect_bids_part(
                    "ses", self.output_path
                )

        if self.ezbids:
            hash_string = helper_functions.hash_fields(**collection_of_fields)
            self.sidecar_template["SeriesDescription"] = hash_string

        if output_path:
            if not isinstance(output_path, pathlib.Path):
                output_path = pathlib.Path(output_path)

            if len(output_path.suffixes) > 1:
                temp_output_path = str(output_path)
                for suffix in output_path.suffixes:
                    temp_output_path = re.sub(suffix, "", temp_output_path)
                output_path = pathlib.Path(temp_output_path).with_suffix(".json")

            with open(output_path, "w") as outfile:
                json.dump(
                    helper_functions.replace_nones(self.sidecar_template),
                    outfile,
                    indent=4,
                )
        else:
            print(
                json.dumps(
                    helper_functions.replace_nones(self.sidecar_template), indent=4
                )
            )

    def write_out_blood_files(
        self, new_file_name_with_entities=None, destination_folder=None
    ):
        recording_entity = "_recording-manual"

        if not new_file_name_with_entities:
            new_file_name_with_entities = pathlib.Path(self.nifti_file)
        if not destination_folder:
            destination_folder = pathlib.Path(self.nifti_file).parent

        if "_pet" in new_file_name_with_entities.name:
            if (
                new_file_name_with_entities.suffix == ".gz"
                and len(new_file_name_with_entities.suffixes) > 1
            ):
                new_file_name_with_entities = new_file_name_with_entities.with_suffix(
                    ""
                ).with_suffix("")

            blood_file_name = new_file_name_with_entities.stem.replace(
                "_pet", recording_entity + "_blood"
            )
        else:
            blood_file_name = (
                new_file_name_with_entities.stem + recording_entity + "_blood"
            )

        if self.spreadsheet_metadata.get("blood_tsv", {}) != {}:
            blood_tsv_data = self.spreadsheet_metadata.get("blood_tsv")
            if type(blood_tsv_data) is pd.DataFrame or type(blood_tsv_data) is dict:
                if type(blood_tsv_data) is dict:
                    blood_tsv_data = pd.DataFrame(blood_tsv_data)
                # write out blood_tsv using pandas csv write
                blood_tsv_data.to_csv(
                    os.path.join(destination_folder, blood_file_name + ".tsv"),
                    sep="\t",
                    index=False,
                )

            elif type(blood_tsv_data) is str:
                # write out with write
                with open(
                    os.path.join(destination_folder, blood_file_name + ".tsv"), "w"
                ) as outfile:
                    outfile.writelines(blood_tsv_data)
            else:
                raise (
                    f"blood_tsv dictionary is incorrect type {type(blood_tsv_data)}, must be type: "
                    f"pandas.DataFrame"
                )

        # if there's blood data in the tsv then write out the sidecar file too
        if (
            self.spreadsheet_metadata.get("blood_json", {}) != {}
            and self.spreadsheet_metadata.get("blood_tsv", {}) != {}
        ):
            blood_json_data = self.spreadsheet_metadata.get("blood_json")
            if type(blood_json_data) is dict:
                # write out to file with json dump
                pass
            elif type(blood_json_data) is str:
                # write out to file with json dumps
                blood_json_data = json.loads(blood_json_data)
            else:
                raise (
                    f"blood_json dictionary is incorrect type {type(blood_json_data)}, must be type: dict or str"
                    f"pandas.DataFrame"
                )

            with open(
                os.path.join(destination_folder, blood_file_name + ".json"), "w"
            ) as outfile:
                json.dump(blood_json_data, outfile, indent=4)

    def update_pet_json(self, pet_json_path):
        """given a json file (or a path ending in .json) update or create a PET json file with information collected
        from an ecat file.
        :param pet_json: a path to a json file
        :type pet_json: str or pathlib.Path
        :return: None
        """

        # open the json file if it exists
        if isinstance(pet_json_path, str):
            pet_json = pathlib.Path(pet_json_path)
        if pet_json.exists():
            with open(pet_json_path, "r") as json_file:
                try:
                    pet_json = json.load(json_file)
                except json.decoder.JSONDecodeError:
                    logger.warning(
                        f"Unable to load json file at {pet_json_path}, skipping."
                    )

            # update the template with values from the json file
            self.sidecar_template.update(pet_json)

        if self.spreadsheet_metadata.get("nifti_json", None):
            self.sidecar_template.update(self.spreadsheet_metadata["nifti_json"])

        self.populate_sidecar(**self.kwargs)
        self.prune_sidecar()

        # check metadata radio inputs
        self.sidecar_template.update(check_meta_radio_inputs(self.sidecar_template))

        self.show_sidecar(output_path=pet_json_path)

    def json_out(self):
        """
        Dumps entire ecat header and header info into stdout formatted as json.

        :return: None
        """
        temp_json = json.dumps(self.ecat_info, indent=4)
        print(temp_json)

    def convert(self):
        """
        Convert ecat to nifti
        :return: None
        """
        self.output_path = pathlib.Path(self.make_nifti())
        self.sidecar_path = self.output_path.parent / self.output_path.stem
        self.sidecar_path = self.sidecar_path.with_suffix(".json")
        self.populate_sidecar(**self.kwargs)
        self.prune_sidecar()
        self.show_sidecar(output_path=self.sidecar_path)
        self.write_out_blood_files()

        if telemetry_enabled:
            send_telemetry(self.telemetry_data)
